"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
This module contains logic to serialize different platforms
"""

from corl.evaluation.runners.section_factories.plugins.platform_serializer import PlatformSerializer


class SerializeCWH3D(PlatformSerializer):
    """
    This is a implementation of the abstract class PlatformSerializer.
    This provides functionality for serializing a 1D docking platform.
    """

    def serialize(self, platform):
        """
        Parameters
        ----------
        platform
            The incoming 2d inspection platform object that needs to be serialized

        Returns
        -------
        dict
            the serialized represnetation of the 1d docking platform.
        """

        dictionary = {
            'name': platform.name,
            'position': platform.position,
            'velocity': platform.velocity,
            'sim_time': platform.sim_time,
            'operable': platform.operable,
        }

        dictionary["controllers"] = {}
        for controller in platform.controllers.values():
            dictionary["controllers"][controller.name] = {"applied_controls": controller.get_applied_control()}

        dictionary["sensors"] = {}
        for sensor in platform.sensors.values():
            dictionary["sensors"][sensor.name] = {"measurement": sensor.get_measurement()}

        return dictionary



"""
--------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core  Extension.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------

This module implements a glue which passes a coordinate axis as an observation to an agent.

Author: Kochise Bennett
"""
import typing
from collections import OrderedDict
from functools import cached_property

import gymnasium
import numpy as np
from corl.glues.base_glue import BaseAgentGlue, BaseAgentGlueValidator
from corl.libraries.property import BoxProp, DictProp
from corl.libraries.units import corl_get_ureg
from pydantic import validator


class CoordinateAxisGlueValidator(BaseAgentGlueValidator):
    """
    A configuration validator for CoordinateAxisGlue

    Attributes
    ----------
    axis : str
        name of the axis to observe
    """
    axis: str = 'x'

    @validator('axis')
    def axis_must_be_in_xyz(cls, v):
        """Validate axis"""
        if v.lower() not in ['x', 'y', 'z']:
            raise ValueError('Axis must be x, y, or z')
        return v


class CoordinateAxisGlue(BaseAgentGlue):
    """Glue that simply outputs a simple 16x16 image for testing purposes"""

    class Fields:
        """
        Fields in this glue
        """
        DIRECT_OBSERVATION = "direct_observation"

    @staticmethod
    def get_validator() -> typing.Type[CoordinateAxisGlueValidator]:
        return CoordinateAxisGlueValidator

    def get_unique_name(self) -> str:
        return "Coordinate_Axis_Glue_" + self.config.axis.upper() + "-Axis"

    @cached_property
    def observation_prop(self):
        prop = BoxProp(low=[-1, -1, -1], high=[1, 1, 1], unit="dimensionless")
        return DictProp(spaces={self.Fields.DIRECT_OBSERVATION: prop})

    @cached_property
    def normalized_observation_space(self) -> typing.Optional[gymnasium.spaces.Space]:
        """
        passthrough property
        """
        return self.observation_space

    @cached_property
    def observation_space(self) -> gymnasium.spaces.Space:
        d = gymnasium.spaces.dict.Dict()
        low = np.array([-1.0, -1.0, -1.0])
        high = np.array([1.0, 1.0, 1.0])
        d.spaces[self.Fields.DIRECT_OBSERVATION] = gymnasium.spaces.Box(low=low, high=high, dtype=np.float32)
        return d

    def get_observation(self, other_obs: OrderedDict, obs_space: OrderedDict, obs_units: OrderedDict):
        if self.config.axis.lower() == 'x':
            out = np.array([1.0, 0.0, 0.0], dtype=np.float32)
        elif self.config.axis.lower() == 'y':
            out = np.array([0.0, 1.0, 0.0], dtype=np.float32)
        elif self.config.axis.lower() == 'z':
            out = np.array([0.0, 0.0, 1.0], dtype=np.float32)

        d = OrderedDict()
        d[self.Fields.DIRECT_OBSERVATION] = corl_get_ureg().Quantity(out, "dimensionless")

        return d



import copy
import typing
import functools
import pettingzoo
import numpy as np
import gymnasium as gym
import scipy.spatial.transform as transform
import safe_autonomy_simulation.sims.inspection as sim
import safe_autonomy_sims.pettingzoo.inspection.reward as r
import safe_autonomy_sims.pettingzoo.inspection.utils as utils


class WeightedSixDofMultiInspectionEnv(pettingzoo.ParallelEnv):
    r"""
    ## Environment

    In this weighted six dof inspection environment, the goal is for a single deputy spacecraft 
    to navigate around and inspect the entire surface of a chief spacecraft.

    The chief is covered in 100 inspection points that the agent must observe
    while they are illuminated by the moving sun. The points are weighted by
    priority, such that it is more important to inspect some points than others.
    A unit vector is used to indicate the direction of highest importance, where
    points are weighted based on their angular distance to this vector. All
    point weights add up to a value of one. The optimal policy will inspect
    points whose cumulative weight exceeds 0.95 within 2 revolutions of the sun
    while using as little fuel as possible.  
    
    In this six DOF inspection environment, the agent controls its translational
    and rotational movement, requiring it to orient itself towards the chief for
    inspection.
    
    __Note: the policy selects a new action every 10 seconds__

    ## Action Space

    Actions are thruster force values for each of the 3 bi-directional thrusters
    on the x-, y-, and z-axis of the deputy spacecraft's body frame and the
    reaction wheel abstraction controlling the deputy's x, y, and z moments with
    scalar values. These controls are able to move and rotate the spacecraft in
    any direction.

    The action space is a `Box` with the following bounds:

    | Index | Action        | Control Min | Control Max | Type (units) |
    |-------|---------------|-------------|-------------|--------------|
    | 0     | x-axis thrust | -1          | 1           | Force (N)    |
    | 1     | y-axis thrust | -1          | 1           | Force (N)    |
    | 2     | z-axis thrust | -1          | 1           | Force (N)    |
    | 3     | x-axis moment | -0.001      | 0.001       | Moment (Nm)  |
    | 4     | y-axis moment | -0.001      | 0.001       | Moment (Nm)  |
    | 5     | z-axis moment | -0.001      | 0.001       | Moment (Nm)  |

    ## Observation Space

    At each timestep, the agent receives the observation,
    $o = [x, y, z, |pos|, |x|, |y|, |z|, v_x, v_y, v_z, |v|, |v_x|, |v_y|, |v_z|, \omega_{x}, \omega_{y}, \omega_{z}, \theta_{x}, \theta_{y}, \theta_{z}, f, \theta_{sun}, n, x_{ups}, y_{ups}, z_{ups}, x_{pv}, y_{pv}, z_{pv}, w_{points}, p_o]$, where:

    * $x, y,$ and $z$ represent the deputy's position relative to the chief,
    * $|pos|, |x|, |y|, |z|$ is the magnorm representation of the deputy's position relative to the chief
    * $v_x, v_y,$ and $v_z$ represent the deputy's directional velocity relative to the chief,
    * $|v|, |v_x|, |v_y|, |v_z|$ is the magnorm representation of the deputy's velocity relative to the chief
    * $\omega_x, \omega_y, \omega_z$ are the components of the deputy's angular velocity
    * $\theta_{cam}$ is the camera's orientation in Hill's frame
    * $\theta_{x}, \theta_{y}, \theta_{z}$ are the deputy axis coordinates in Hill's frame
    * $f$ is the dot-product between the camera orientation vector and the relative position between the deputy and the chief. This value is 1 when the camera is pointing at the chief.  
    * $\theta_{sun}$ is the angle of the sun,
    * $n$ is the number of points that have been inspected so far and,
    * $x_{ups}, y_{ups},$ and $z_{ups}$ are the unit vector elements pointing to the nearest large cluster of unispected points as determined by the *Uninspected Points Sensor*.
    * $x_{pv}, y_{pv},$ and $z_{pv}$ are the unit vector elements pointing to the priority vector indicating point priority.
    * $w_{points}$ is the cumulative weight of inpsected points
    * $p_o$ is the dot-product between the uninspected points cluster given by the Uninspected Points Sensor and the deputy's position. This signals if the uninspected points are occluded for if the camera is facing the points but they are not being inspected, there is occlusion.

    The observation space is a `Box` with the following bounds:

    | Index | Observation                                                 | Min | Max | Type (units) |
    |-------|-------------------------------------------------------------|-----|-----|--------------|
    | 0     | x position of the deputy in Hill's frame                    | -inf| inf | Position (m) |
    | 1     | y position of the deputy in Hill's frame                    | -inf| inf | Position (m) |
    | 2     | z position of the deputy in Hill's frame                    | -inf| inf | Position (m) |
    | 3     | |pos| magnitude of the deputy's position                    | -inf| inf | Position (m) |
    | 4     | |x| magnitude of the deputy's x position                    | -inf| inf | Position (m) |
    | 5     | |y| magnitude of the deputy's y position                    | -inf| inf | Position (m) |
    | 6     | |z| magnitude of the deputy's z position                    | -inf| inf | Position (m) |
    | 7     | x component of the deputy's velocity                        | -inf| inf | Velocity (m/s) |
    | 8     | y component of the deputy's velocity                        | -inf| inf | Velocity (m/s) |
    | 9     | z component of the deputy's velocity                        | -inf| inf | Velocity (m/s) |
    | 10    | |v| magnitude of the deputy's velocity                      | -inf| inf | Velocity (m/s) |
    | 11    | |v_x| magnitude of the x component of the deputy's velocity | -inf| inf | Velocity (m/s) |
    | 12    | |v_y| magnitude of the y component of the deputy's velocity | -inf| inf | Velocity (m/s) |
    | 13    | |v_z| magnitude of the z component of the deputy's velocity | -inf| inf | Velocity (m/s) |
    | 14    | x component of the deputy's angular velocity                | -inf| inf | Angular Velocity (rad/s) |
    | 15    | y component of the deputy's angular velocity                | -inf| inf | Angular Velocity (rad/s) |
    | 16    | z component of the deputy's angular velocity                | -inf| inf | Angular Velocity (rad/s) |
    | 17    | x component of the deputy's orientation                     | 0   | 2pi | Orientation (rad) |
    | 18    | y component of the deputy's orientation                     | 0   | 2pi | Orientation (rad) |
    | 19    | z component of the deputy's orientation                     | 0   | 2pi | Orientation (rad) |
    | 20    | facing chief dot product                                    | -1  | 1   | Scalar |
    | 21    | sun angle                                                   | 0   | 2pi | Angle (rad) |
    | 22    | number of inspected points                                  | 0   | 100 | Scalar |
    | 23    | x component of unit vector pointing to the nearest cluster  | -1  | 1   | Scalar |
    | 24    | y component of unit vector pointing to the nearest cluster  | -1  | 1   | Scalar |
    | 25    | z component of unit vector pointing to the nearest cluster  | -1  | 1   | Scalar |
    | 26    | x component of unit vector pointing to the priority vector  | -1  | 1   | Scalar |
    | 27    | y component of unit vector pointing to the priority vector  | -1  | 1   | Scalar |
    | 28    | z component of unit vector pointing to the priority vector  | -1  | 1   | Scalar |
    | 29    | cumulative weight of inspected points                       | 0   | 1   | Scalar |
    | 30    | dot product between nearest cluster and deputy's position   | -1  | 1   | Scalar |

    ## State Transition Dynamics

     The relative motion between the deputy and chief is represented by
    linearized Clohessy-Wiltshire equations [[1]](#1), given by

    $$
        \dot{\boldsymbol{x}} = A {\boldsymbol{x}} + B\boldsymbol{u},
    $$

    where:
    * state $\boldsymbol{x}=[x,y,z,\dot{x},\dot{y},\dot{z}]^T \in \mathcal{X}=\mathbb{R}^6$
    * control $\boldsymbol{u}= [F_x,F_y,F_z]^T \in \mathcal{U} = [-1N, 1N]^3$

    * $$
        A =
    \begin{bmatrix}
    0 & 0 & 0 & 1 & 0 & 0 \\
    0 & 0 & 0 & 0 & 1 & 0 \\
    0 & 0 & 0 & 0 & 0 & 1 \\
    3n^2 & 0 & 0 & 0 & 2n & 0 \\
    0 & 0 & 0 & -2n & 0 & 0 \\
    0 & 0 & -n^2 & 0 & 0 & 0 \\
    \end{bmatrix},
        B =
    \begin{bmatrix}
    0 & 0 & 0 \\
    0 & 0 & 0 \\
    0 & 0 & 0 \\
    \frac{1}{m} & 0 & 0 \\
    0 & \frac{1}{m} & 0 \\
    0 & 0 & \frac{1}{m} \\
    \end{bmatrix},
    $$

    * mean motion constant $n = 0.001027 rad/s$

    The body frame rotational motion state transition of each spacecraft
    given its quaternion orientation and angular velocity
    $[q_1, q_2, q_3, q_4, \omega_x, \omega_y, \omega_z]$ is defined by

    $$
    \begin{bmatrix}
    \dot{q_1} \\
    \dot{q_2} \\
    \dot{q_3} \\
    \dot{q_4} \\
    \dot{\omega_x} \\
    \dot{\omega_y} \\
    \dot{\omega_z} \\
    \end{bmatrix}
    =
    \begin{bmatrix}
    \frac{1}{2}(q_4\omega_x - q_3\omega_y + q_2\omega_z) \\
    \frac{1}{2}(q_3\omega_x + q_4\omega_y - q_1\omega_z) \\
    \frac{1}{2}(-q_2\omega_x + q_1\omega_y + q_4\omega_z) \\
    \frac{1}{2}(-q_1\omega_x - q_2\omega_y - q_3\omega_z) \\
    J_1^{-1}((J_2 - J_3)\omega_y\omega_z) \\
    J_2^{-1}((J_3 - J_1)\omega_x\omega_z) \\
    J_3^{-1}((J_1 - J_2)\omega_x\omega_y) \\
    \end{bmatrix}
    $$

    where

    $$
    J =
    \begin{bmatrix}
    0.0573 & 0.0 & 0.0 \\
    0.0 & 0.0573 & 0.0 \\
    0.0 & 0.0 & 0.0573
    \end{bmatrix}
    $$
    
    is an inertial matrix.

    ## Rewards

    The reward $r_t$ at each time step is the sum of the following terms:
    * $r_t += 1.0(weight\_inspected\_points_t - weight\_inspected\_points_{t-1})$.
        * a dense reward for inspecting new points
    * $r_t += 1$ if $weight\_inspected\_points_i \geq 0.95$ and $FFT_radius \geq crash\_region\_radius$, $r = -1$ if $weight\_inspected\_points_i \geq 0.95$ and $FFT_radius < crash\_region\_radius$, else 0
        * a sparse reward for successfully inspecting the chief
        * positive reward for successful inspection and safe trajectory
        * negative reward for successful inspection and unsafe trajectory (FFT collision with chief)
    * $r_t += -1$ if $radius < crash\_region\_radius$, else 0
        * a sparse reward punishing the agent for crashing with the chief
    * $r_t += -1$ if $||pos_{chief} - pos_{deputy}|| > distance_{max}$, else 0.
        * a sparse reward punishing the agent for moving too far from the chief
    * $r_t += 0.001$ if $t < t_{max}$, else 0
        * a dense reward for staying alive
    * $r_t += 0.0005 * e^{-|\delta_t(f, 1)^2 / \epsilon|}$
        * a dense reward for facing the chief
        * $\delta_t(f, 1)$ is the difference between the $f$, dot-product between the camera orientation vector and the relative position between the deputy and the chief, and 1.
        * $\epsilon$ is the length of the reward curve for the exponential decay (configurable)
    * $r_t += -0.1||\boldsymbol{u}||$
        * a dense reward for minimizing thruster usage (fuel cost)

    ## Starting State

    At the start of an episode, the state is randomly initialized with the following conditions:

    * chief $(x,y,z)$ = $(0, 0, 0)$
    * chief radius = $10 m$
    * chief # of points = $100$
    * deputy position $(x, y, z)$ is converted after randomly selecting the position in polar notation $(r, \phi, \psi)$ using a uniform distribution with
        * $r \in [50, 100] m$
        * $\psi \in [0, 2\pi] rad$
        * $\phi \in [-\pi/2, \pi/2] rad$
        * $x = r \cos{\psi} \cos{\phi}$
        * $y = r \sin{\psi} \cos{\phi}$
        * $z = r \sin{\phi}$
    * deputy $(v_x, v_y, v_z)$ is converted after randomly selecting the velocity in polar notation $(r, \phi, \psi)$ using a Gaussian distribution with
        * $v \in [0, 0.3]$ m/s
        * $\psi \in [0, 2\pi] rad$
        * $\phi \in [-\pi/2, \pi/2] rad$
        * $v_x = v \cos{\psi} \cos{\phi}$
        * $v_y = v \sin{\psi} \cos{\phi}$
        * $v_z = v \sin{\phi}$
    * deputy $(\omega_x, \omega_y, \omega_z)$ is sampled from a uniform distribution between $[-0.01, -0.01, -0.01]$ rad/s and $[0.01, 0.01, 0.01]$ rad/s
    * Initial sun angle is randomly selected using a uniform distribution
        * $\theta_{sun} \in [0, 2\pi] rad$
        * If the deputy is initialized where it's sensor points within 60 degrees of the sun, its position is negated such that the sensor points away from the sun.

    ## Episode End

    An episode will terminate if any of the following conditions are met:

    * Termination: the agent exceeds a `max_distance = 800` meter radius away from the chief
    * Termination: the agent moves within a `crash_region_radius = 15` meter radius around the chief
    * Termination: the cumulative weight of inspected points exceeds 0.95
    * Truncation: the maximum number of timesteps, `max_timesteps = 1224`

    The episode is considered done and successful if and only if the cumulative weight of
    inspected points exceeds 0.95 while the deputy remains on a safe trajectory
    (not on a collision course with the chief).

    ## References

    <a id="1">[1]</a>
    Clohessy, W., and Wiltshire, R., “Terminal Guidance System for Satellite Rendezvous,” *Journal of the Aerospace Sciences*, Vol. 27, No. 9, 1960, pp. 653–658.

    <a id="2">[2]</a>
    Dunlap, K., Mote, M., Delsing, K., and Hobbs, K. L., “Run Time Assured Reinforcement Learning for Safe Satellite
    Docking,” *Journal of Aerospace Information Systems*, Vol. 20, No. 1, 2023, pp. 25–36. [https://doi.org/10.2514/1.I011126](https://doi.org/10.2514/1.I011126).

    <a id="3">[3]</a>
    Gaudet, B., Linares, R., and Furfaro, R., “Adaptive Guidance and Integrated Navigation with Reinforcement Meta-Learning,”
    *CoRR*, Vol. abs/1904.09865, 2019. URL [http://arxiv.org/abs/1904.09865](http://arxiv.org/abs/1904.09865).

    <a id="4">[4]</a>
    Battin, R. H., “An introduction to the mathematics and methods of astrodynamics,” 1987.

    <a id="5">[5]</a>
    Campbell, T., Furfaro, R., Linares, R., and Gaylor, D., “A Deep Learning Approach For Optical Autonomous Planetary Relative
    Terrain Navigation,” 2017.

    <a id="6">[6]</a>
    Furfaro, R., Bloise, I., Orlandelli, M., Di Lizia, P., Topputo, F., and Linares, R., “Deep Learning for Autonomous Lunar
    Landing,” 2018.

    <a id="7">[7]</a>
    Lei, H. H., Shubert, M., Damron, N., Lang, K., and Phillips, S., “Deep reinforcement Learning for Multi-agent Autonomous
    Satellite Inspection,” *AAS Guidance Navigation and Control Conference*, 2022.

    <a id="8">[8]</a>
    Aurand, J., Lei, H., Cutlip, S., Lang, K., and Phillips, S., “Exposure-Based Multi-Agent Inspection of a Tumbling Target Using
    Deep Reinforcement Learning,” *AAS Guidance Navigation and Control Conference*, 2023.

    <a id="9">[9]</a>
    Brandonisio, A., Lavagna, M., and Guzzetti, D., “Reinforcement Learning for Uncooperative Space Objects Smart Imaging
    Path-Planning,” The *Journal of the Astronautical Sciences*, Vol. 68, No. 4, 2021, pp. 1145–1169. [https://doi.org/10.1007/s40295-021-00288-7](https://doi.org/10.1007/s40295-021-00288-7).
    """

    def __init__(
        self,
        num_agents: int = 2,
        success_threshold: float = 0.95,
        crash_radius: float = 15,
        max_distance: float = 800,
        max_time: float = 1000,
    ) -> None:
        self.possible_agents = [f"deputy_{i}" for i in range(num_agents)]

        # Environment parameters
        self.crash_radius = crash_radius
        self.max_distance = max_distance
        self.max_time = max_time
        self.success_threshold = success_threshold

        # Episode level information
        self.prev_state = None
        self.prev_num_inspected = 0
        self.prev_weight_inspected = 0.0

    def reset(
        self, seed: int | None = None, options: dict[str, typing.Any] | None = None
    ) -> tuple[typing.Any, dict[str, typing.Any]]:
        self.agents = copy.copy(self.possible_agents)
        self.rng = np.random.default_rng(seed)
        self._init_sim()  # sim is light enough we just reconstruct it
        self.simulator.reset()
        observations = {a: self._get_obs(a) for a in self.agents}
        infos = {a: self._get_info(a) for a in self.agents}
        self.prev_state = None
        self.prev_num_inspected = 0
        self.prev_weight_inspected = 0.0
        return observations, infos

    def step(
        self, actions: dict[str, typing.Any]
    ) -> tuple[typing.Any, typing.SupportsFloat, bool, bool, dict[str, typing.Any]]:
        # Store previous simulator state
        self.prev_state = self.sim_state.copy()
        self.prev_num_inspected = (
            self.chief.inspection_points.get_num_points_inspected()
        )
        self.prev_weight_inspected = (
            self.chief.inspection_points.get_total_weight_inspected()
        )

        # Update simulator state
        for agent, action in actions.items():
            self.deputies[agent].add_control(action)
        self.simulator.step()

        # Get info from simulator
        observations = {a: self._get_obs(a) for a in self.agents}
        rewards = {a: self._get_reward(a) for a in self.agents}
        infos = {a: self._get_info(a) for a in self.agents}
        terminations = {a: self._get_terminated(a) for a in self.agents}
        truncations = {
            a: False for a in self.agents
        }  # used to signal episode ended unexpectedly

        # End episode if any agent is terminated or truncated
        if any(terminations.values() or any(truncations.values())):
            truncations = {a: True for a in self.agents}
            terminations = {a: True for a in self.agents}
            self.agents = []

        return observations, rewards, terminations, truncations, infos

    def _init_sim(self):
        # Initialize spacecraft, sun, and simulator
        priority_vector = self.rng.uniform(-1, 1, size=3)
        priority_vector /= np.linalg.norm(priority_vector)  # convert to unit vector
        self.chief = sim.SixDOFTarget(
            name="chief",
            num_points=100,
            radius=1,
            priority_vector=priority_vector,
        )
        self.deputies = {
            a: sim.SixDOFInspector(
                name=a,
                position=utils.polar_to_cartesian(
                    r=self.rng.uniform(50, 100),
                    phi=self.rng.uniform(-np.pi / 2, np.pi / 2),
                    theta=self.rng.uniform(0, 2 * np.pi),
                ),
                velocity=utils.polar_to_cartesian(
                    r=self.rng.uniform(0, 0.3),
                    phi=self.rng.uniform(-np.pi / 2, np.pi / 2),
                    theta=self.rng.uniform(0, 2 * np.pi),
                ),
            )
            for a in self.agents
        }
        self.sun = sim.Sun(theta=self.rng.uniform(0, 2 * np.pi))
        self.simulator = sim.InspectionSimulator(
            frame_rate=10,
            inspectors=list(self.deputies.values()),
            targets=[self.chief],
            sun=self.sun,
        )

    def _get_obs(self, agent: typing.Any) -> np.ndarray:
        deputy = self.deputies[agent]
        obs = self.observation_space(agent).sample()
        obs[:3] = deputy.position
        obs[3] = np.linalg.norm(deputy.position)
        obs[4:7] = np.abs(deputy.position)
        obs[7:10] = deputy.velocity
        obs[10] = np.linalg.norm(deputy.velocity)
        obs[11:14] = np.abs(deputy.velocity)
        obs[14:17] = deputy.angular_velocity
        obs[17:20] = transform.Rotation.from_quat(deputy.orientation).as_euler("XYZ")
        obs[20] = np.dot(
            transform.Rotation.from_quat(deputy.camera.orientation).as_euler("XYZ"),
            (self.chief.position - deputy.position)
            / np.linalg.norm(self.chief.position - deputy.position),
        )
        obs[21] = self.sun.theta
        obs[22] = self.chief.inspection_points.get_num_points_inspected()
        obs[23:26] = self.chief.inspection_points.kmeans_find_nearest_cluster(
            camera=deputy.camera, sun=self.sun
        )
        obs[26:29] = self.chief.inspection_points.priority_vector
        obs[29] = self.chief.inspection_points.get_total_weight_inspected()
        obs[30] = np.dot(
            transform.Rotation.from_quat(deputy.camera.orientation).as_euler("XYZ"),
            self.chief.inspection_points.kmeans_find_nearest_cluster(
                camera=deputy.camera, sun=self.sun
            ),
        )
        return obs

    def _get_info(self, agent: typing.Any) -> dict[str, typing.Any]:
        return {}

    def _get_reward(self, agent: typing.Any) -> float:
        reward = 0
        deputy = self.deputies[agent]

        # Dense rewards
        reward += r.weighted_observed_points_reward(
            chief=self.chief, prev_weight_inspected=self.prev_weight_inspected
        )
        reward += r.delta_v_reward(
            v=deputy.velocity, prev_v=self.prev_state[agent][3:6]
        )

        reward += r.live_timestep_reward(t=self.simulator.sim_time, t_max=self.max_time)
        reward += r.facing_chief_reward(chief=self.chief, deputy=deputy, epsilon=0.01)

        # Sparse rewards
        reward += (
            r.weighted_inspection_success_reward(
                chief=self.chief, total_weight=self.success_threshold
            )
            if utils.closest_fft_distance(chief=self.chief, deputy=deputy)
            < self.crash_radius
            else -1.0
        )
        reward += r.crash_reward(
            chief=self.chief, deputy=deputy, crash_radius=self.crash_radius
        )

        return reward

    def _get_terminated(self, agent: typing.Any) -> bool:
        deputy = self.deputies[agent]

        # Get state info
        d = utils.rel_dist(pos1=self.chief.position, pos2=deputy.position)

        # Determine if in terminal state
        oob = d > self.max_distance
        crash = d < self.crash_radius
        timeout = self.simulator.sim_time > self.max_time
        all_inspected = self.prev_weight_inspected >= self.success_threshold

        return oob or crash or timeout or all_inspected

    @functools.lru_cache(maxsize=None)
    def observation_space(self, agent: typing.Any) -> gym.Space:
        return gym.spaces.Box(
            np.concatenate(
                (
                    [-np.inf] * 3,  # deputy position
                    [-np.inf] * 4,  # deputy position magnorm
                    [-np.inf] * 3,  # deputy velocity
                    [-np.inf] * 4,  # deputy velocity magnorm
                    [-np.inf] * 3,  # deputy angular velocity
                    [-2 * np.pi] * 3,  # deputy orientation (euler)
                    [-1],  # facing chief dot product
                    [0],  # sun angle
                    [0],  # number of inspected points
                    [-1] * 3,  # nearest cluster unit vector
                    [-1] * 3,  # priority vector unit vector
                    [0],  # cumulative weight of inspected points
                    [-1],  # facing cluster dot product
                )
            ),
            np.concatenate(
                (
                    [np.inf] * 3,  # deputy position
                    [np.inf] * 4,  # deputy position magnorm
                    [np.inf] * 3,  # deputy velocity
                    [np.inf] * 4,  # deputy velocity magnorm
                    [np.inf] * 3,  # deputy angular velocity
                    [2 * np.pi] * 3,  # deputy orientation
                    [1],  # facing chief dot product
                    [2 * np.pi],  # sun angle
                    [100],  # number of inspected points
                    [1] * 3,  # nearest cluster unit vector
                    [1] * 3,  # priority vector unit vector
                    [1],  # cumulative weight of inspected points
                    [1],  # facing cluster dot product
                )
            ),
            shape=(31,),
            dtype=np.float64,
        )

    @functools.lru_cache(maxsize=None)
    def action_space(self, agent: typing.Any) -> gym.Space:
        return gym.spaces.Box(
            np.array([-1, -1, -1, -0.001, -0.001, -0.001]),
            np.array([1, 1, 1, 0.001, 0.001, 0.001]),
            shape=(6,),
            dtype=np.float64,
        )

    @property
    def sim_state(self) -> dict:
        state = {
            "chief": self.chief.state,
        }
        for a in self.agents:
            state[a] = self.deputies[a].state
        return state



"""
--------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core  Extension.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------

This module implements an extension of the CoRL Magnitude Glue with the observation_prop property.
"""

from functools import cached_property

from corl.glues.common.magnitude import MagnitudeGlue
from corl.libraries.property import BoxProp, DictProp


class SimsMagnitudeGlue(MagnitudeGlue):

    @cached_property
    def observation_prop(self):
        prop = DictProp(name=self._uname, spaces={self.Fields.MAG: BoxProp(low=[-10000], high=[10000], unit='m/s')})
        return prop



"""
--------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core  Extension.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------

This module implements a common platform object for the safe autonomy environments.
"""

from corl.simulators.base_platform import BasePlatform


class BaseSafeRLPlatform(BasePlatform):
    """
    Base platform for safe autonomy sims
    Allows for saving an action to the platform for when the platform needs
    to give an action to the environment during the environment step function.

    Parameters
    ----------
    platform_name : str
        Name of the platform.
    platform : sim_entity
        Backend simulation entity associated with the platform.
    parts_list : dict
        Platform-specific parts configuration dictionary.
    sim_time : float
        simulation time at platform creation
    """

    def __init__(self, platform_name, platform, parts_list, sim_time=0.0):  # pylint: disable=W0613
        super().__init__(platform_name=platform_name, platform=platform, parts_list=parts_list)
        self._sim_time = sim_time

    @property
    def sim_time(self):
        """
        The current simulation time in seconds.

        Returns
        -------
        float
            Current simulation time.
        """
        return self._sim_time

    @sim_time.setter
    def sim_time(self, time):
        self._sim_time = time



"""
--------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core  Extension.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------

Wrapper glue which returns a velocity constraint based on position and velocity sensor data as an observation.
"""
import typing
from collections import OrderedDict
from functools import cached_property

import gymnasium
import numpy as np
from corl.glues.base_glue import BaseAgentGlueNormalizationValidator
from corl.glues.common.observe_sensor import ObserveSensor, ObserveSensorValidator
from corl.libraries.normalization import StandardNormalNormalizer
from corl.libraries.property import BoxProp, DictProp
from corl.libraries.units import corl_get_ureg


class VelocityLimitGlueValidator(ObserveSensorValidator):
    """
    A configuration validator for the VelocityLimitGlue.

    Attributes
    ----------
    velocity_threshold: float
        The maximum tolerated velocity within docking region without crashing
    threshold_distance: float
        The radius of the docking region
    mean_motion: float
        Orbital mean motion of Hill's reference frame's circular orbit in rad/s
    slope: float
        The slope of the linear velocity limit as a function of distance from docking region
    normalization: BaseAgentGlueNormalizationValidator
        Default normalization
    """
    velocity_threshold: float = 0.2
    threshold_distance: float = 0.5
    mean_motion: float = 0.001027
    slope: float = 2.0
    normalization: BaseAgentGlueNormalizationValidator = BaseAgentGlueNormalizationValidator(normalizer=StandardNormalNormalizer)


class VelocityLimitGlue(ObserveSensor):
    """
    Computes a velocity constraint from position and velocity sensor data.
    """

    @staticmethod
    def get_validator():
        return VelocityLimitGlueValidator

    def get_unique_name(self) -> str:
        return super().get_unique_name() + "_VelocityLimit"

    def _vel_limit(self, dist):
        return self.config.velocity_threshold + (self.config.slope * self.config.mean_motion * (dist - self.config.threshold_distance))

    @cached_property
    def observation_prop(self):
        prop = BoxProp(low=[-10000], high=[10000], unit='m/s')
        return DictProp(spaces={self.Fields.DIRECT_OBSERVATION: prop})

    @cached_property
    def observation_space(self) -> typing.Optional[gymnasium.spaces.Space]:
        pos_obs_space = super().observation_space[self.Fields.DIRECT_OBSERVATION]
        high = self._vel_limit(np.linalg.norm(pos_obs_space.high)) * np.sign(pos_obs_space.high)[0]
        low = self._vel_limit(np.linalg.norm(pos_obs_space.low)) * np.sign(pos_obs_space.low)[0]
        d = gymnasium.spaces.dict.Dict()
        d.spaces[self.Fields.DIRECT_OBSERVATION] = gymnasium.spaces.Box(low, high, shape=(1, ), dtype=np.float32)
        return d

    def get_observation(self, other_obs: OrderedDict, obs_space: OrderedDict, obs_units: OrderedDict):
        pos_obs = super().get_observation(other_obs, obs_space, obs_units)[self.Fields.DIRECT_OBSERVATION]
        obs = corl_get_ureg().Quantity(np.array([self._vel_limit(np.linalg.norm(pos_obs.m))], dtype=np.float32), str(pos_obs.units))
        d = OrderedDict()
        d[self.Fields.DIRECT_OBSERVATION] = obs
        return d

    @cached_property
    def normalized_action_space(self) -> typing.Optional[gymnasium.spaces.Space]:
        return self.action_space

    @cached_property
    def normalized_observation_space(self) -> typing.Optional[gymnasium.spaces.Space]:
        return self.observation_space



import typing
import numpy as np
import gymnasium as gym
import safe_autonomy_simulation.sims.inspection as sim
import safe_autonomy_sims.gym.inspection.reward as r
import safe_autonomy_sims.gym.inspection.utils as utils


class WeightedInspectionEnv(gym.Env):
    r"""
    In this weighted inspection environment, the goal is for a single deputy spacecraft 
    to navigate around and inspect the entire surface of a chief spacecraft.
 
    The chief is covered in 100 inspection points that the agent must observe
    while they are illuminated by the moving sun. The points are weighted by
    priority, such that it is more important to inspect some points than others.
    A unit vector is used to indicate the direction of highest importance, where
    points are weighted based on their angular distance to this vector. All
    point weights add up to a value of one. The optimal policy will inspect
    points whose cumulative weight exceeds 0.95 within 2 revolutions of the sun
    while using as little fuel as possible.
    
    In this weighted inspection environment, the agent only controls its
    translational motion and is always assumed to be pointing at the chief spacecraft.
    
    __Note: the policy selects a new action every 10 seconds__

    ## Action Space

    Actions are thruster force values for each of the 3 bi-directional thrusters
    on the x-, y-, and z-axis of the deputy spacecraft's body frame.

    The action space is a `Box(-1, 1, shape=(3,))` where each value in the array
    represents the force applied to the x-, y-, and z-axis of the deputy spacecraft.

    | Index | Action        | Control Min | Control Max | Type (units) |
    |-------|---------------|-------------|-------------|--------------|
    | 0     | x-axis thrust | -1          | 1           | Force (N)    |
    | 1     | y-axis thrust | -1          | 1           | Force (N)    |
    | 2     | z-axis thrust | -1          | 1           | Force (N)    |

    ## Observation Space

    At each timestep, the agent receives the observation,
    $o = [x, y, z, v_x, v_y, v_z, \theta_{sun}, n, x_{ups}, y_{ups}, z_{ups}, x_{pv}, y_{pv}, z_{pv}, w_{points}]$, where:

    * $x, y,$ and $z$ represent the deputy's position in the Hill's frame,
    * $v_x, v_y,$ and $v_z$ represent the deputy's directional velocity in the Hill's frame,
    * $\theta_{sun}$ is the angle of the sun,
    * $n$ is the number of points that have been inspected so far
    * $x_{ups}, y_{ups},$ and $z_{ups}$ are the unit vector elements pointing to the nearest large cluster of unispected points
    * $x_{pv}, y_{pv},$ and $z_{pv}$ are the unit vector elements pointing to the priority vector indicating point priority
    * $w_{points}$ is the cumulative weight of inpsected points

    The observation space is a `Box` with the following bounds:

    | Index | Observation                                            | Min  | Max    | Type (units)   |
    |-------|--------------------------------------------------------|------|--------|----------------|
    | 0     | x position of the deputy in Hill's frame               | -inf | inf    | position (m)   |
    | 1     | y position of the deputy in Hill's frame               | -inf | inf    | position (m)   |
    | 2     | z position of the deputy in Hill's frame               | -inf | inf    | position (m)   |
    | 3     | x velocity of the deputy in Hill's frame               | -inf | inf    | velocity (m/s) |
    | 4     | y velocity of the deputy in Hill's frame               | -inf | inf    | velocity (m/s) |
    | 5     | z velocity of the deputy in Hill's frame               | -inf | inf    | velocity (m/s) |
    | 6     | angle of the sun                                       | 0    | $2\pi$ | angle (rad)    |
    | 7     | number of points inspected                             | 0    | 100    | count          |
    | 8     | x component of unit vector pointing to nearest cluster | -1   | 1      | scalar         |
    | 9     | y component of unit vector pointing to nearest cluster | -1   | 1      | scalar         |
    | 10    | z component of unit vector pointing to nearest cluster | -1   | 1      | scalar         |
    | 11    | x component of unit vector pointing to priority vector | -1   | 1      | scalar         |
    | 12    | y component of unit vector pointing to priority vector | -1   | 1      | scalar         |
    | 13    | z component of unit vector pointing to priority vector | -1   | 1      | scalar         |
    | 14    | cumulative weight of inspected points                  | 0    | 1      | scalar         |

    ## State Transition Dynamics

    The relative motion between the deputy and chief is represented by
    linearized Clohessy-Wiltshire equations [[1]](#1), given by

    $$
        \dot{\boldsymbol{x}} = A {\boldsymbol{x}} + B\boldsymbol{u},
    $$

    where:

    * state $\boldsymbol{x}=[x,y,z,\dot{x},\dot{y},\dot{z}]^T \in \mathcal{X}=\mathbb{R}^6$
    * control $\boldsymbol{u}= [F_x,F_y,F_z]^T \in \mathcal{U} = [-1N, 1N]^3$

    * $$
        A =
    \begin{bmatrix}
    0 & 0 & 0 & 1 & 0 & 0 \\
    0 & 0 & 0 & 0 & 1 & 0 \\
    0 & 0 & 0 & 0 & 0 & 1 \\
    3n^2 & 0 & 0 & 0 & 2n & 0 \\
    0 & 0 & 0 & -2n & 0 & 0 \\
    0 & 0 & -n^2 & 0 & 0 & 0 \\
    \end{bmatrix},
        B =
    \begin{bmatrix}
    0 & 0 & 0 \\
    0 & 0 & 0 \\
    0 & 0 & 0 \\
    \frac{1}{m} & 0 & 0 \\
    0 & \frac{1}{m} & 0 \\
    0 & 0 & \frac{1}{m} \\
    \end{bmatrix},
    $$

    * mean motion constant $n = 0.001027 rad/s$

    ## Rewards

    The reward $r_t$ at each time step is the sum of the following terms:

    * $r_t += 1.0(weight\_inspected\_points_t - weight\_inspected\_points_{t-1})$
        * a dense reward for observing inspection points
    * $r += 1$ if $weight\_inspected\_points_i \geq 0.95$ and $FFT_radius \geq crash\_region\_radius$, $r = -1$ if $weight\_inspected\_points_i \geq 0.95$ and $FFT_radius < crash\_region\_radius$, else 0
        * a sparse reward for successfully inspecting the chief spacecraft
        * positive reward if the agent inspects 95% of the points and is on a safe trajectory
        * negative reward if the agent inspects 95% of the points and is on a collision course with the chief spacecraft
    * $r += -1$ if $radius < crash\_region\_radius$, else 0
        * a sparse reward punishing the agent for crashing with the chief spacecraft
    * $r += -0.1||\boldsymbol{u}||$
        * a dense reward for minimizing fuel usage
        * fuel use is represented by the norm of the control input (action)

    ## Starting State

    At the start of an episode, the state is randomly initialized with the following conditions:

    * chief $(x,y,z)$ = $(0, 0, 0)$
    * chief radius = $10 m$
    * chief # of points = $100$
    * priority unit vector orientation for point weighting is randomly sampled from a uniform distribution using polar notation $(\phi, \psi)$
        * $\psi \in [0, 2\pi] rad$
        * $\phi \in [-\pi/2, \pi/2] rad$
    * deputy position $(x, y, z)$ is converted after randomly selecting the position in polar notation $(r, \phi, \psi)$ using a uniform distribution with
        * $r \in [50, 100] m$
        * $\psi \in [0, 2\pi] rad$
        * $\phi \in [-\pi/2, \pi/2] rad$
        * $x = r \cos{\psi} \cos{\phi}$
        * $y = r \sin{\psi} \cos{\phi}$
        * $z = r \sin{\phi}$
    * deputy $(v_x, v_y, v_z)$ is converted after randomly selecting the velocity in polar notation $(r, \phi, \psi)$ using a Gaussian distribution with
        * $v \in [0, 0.3]$ m/s
        * $\psi \in [0, 2\pi] rad$
        * $\phi \in [-\pi/2, \pi/2] rad$
        * $v_x = v \cos{\psi} \cos{\phi}$
        * $v_y = v \sin{\psi} \cos{\phi}$
        * $v_z = v \sin{\phi}$
    * deputy camera parameters:
        * field of view = $\pi$ rad
        * focal length = $1 m$
    * Initial sun angle is randomly selected using a uniform distribution
        * $\theta_{sun} \in [0, 2\pi] rad$
    * Simulator frame rate = $0.1 Hz$

    ## Episode End

    An episode will end if any of the following conditions are met:

    * Terminated: the agent moves within a `crash_region_radius = 10` meter radius around the chief
    * Terminated: the cumulative weight of inspected points exceeds 0.95
    * Truncated: the maximum number of timesteps, `max_timesteps = 12236`
    * Truncated: the agent exceeds a `max_distance = 800` meter radius away from the chief

    The episode is considered done and successful if and only if the cumulative weight of inspected points exceeds 0.95.

    ## References

    <a id="1">[1]</a>
    Clohessy, W., and Wiltshire, R., “Terminal Guidance System for Satellite Rendezvous,” *Journal of the Aerospace Sciences*, Vol. 27, No. 9, 1960, pp. 653–658.

    <a id="2">[2]</a>
    Dunlap, K., Mote, M., Delsing, K., and Hobbs, K. L., “Run Time Assured Reinforcement Learning for Safe Satellite
    Docking,” *Journal of Aerospace Information Systems*, Vol. 20, No. 1, 2023, pp. 25–36. [https://doi.org/10.2514/1.I011126](https://doi.org/10.2514/1.I011126).

    <a id="3">[3]</a>
    Gaudet, B., Linares, R., and Furfaro, R., “Adaptive Guidance and Integrated Navigation with Reinforcement Meta-Learning,”
    *CoRR*, Vol. abs/1904.09865, 2019. URL [http://arxiv.org/abs/1904.09865](http://arxiv.org/abs/1904.09865).

    <a id="4">[4]</a>
    Battin, R. H., “An introduction to the mathematics and methods of astrodynamics,” 1987.

    <a id="5">[5]</a>
    Campbell, T., Furfaro, R., Linares, R., and Gaylor, D., “A Deep Learning Approach For Optical Autonomous Planetary Relative
    Terrain Navigation,” 2017.

    <a id="6">[6]</a>
    Furfaro, R., Bloise, I., Orlandelli, M., Di Lizia, P., Topputo, F., and Linares, R., “Deep Learning for Autonomous Lunar
    Landing,” 2018.

    <a id="7">[7]</a>
    Lei, H. H., Shubert, M., Damron, N., Lang, K., and Phillips, S., “Deep reinforcement Learning for Multi-agent Autonomous
    Satellite Inspection,” *AAS Guidance Navigation and Control Conference*, 2022.

    <a id="8">[8]</a>
    Aurand, J., Lei, H., Cutlip, S., Lang, K., and Phillips, S., “Exposure-Based Multi-Agent Inspection of a Tumbling Target Using
    Deep Reinforcement Learning,” *AAS Guidance Navigation and Control Conference*, 2023.

    <a id="9">[9]</a>
    Brandonisio, A., Lavagna, M., and Guzzetti, D., “Reinforcement Learning for Uncooperative Space Objects Smart Imaging
    Path-Planning,” The *Journal of the Astronautical Sciences*, Vol. 68, No. 4, 2021, pp. 1145–1169. [https://doi.org/10.1007/s40295-021-00288-7](https://doi.org/10.1007/s40295-021-00288-7).
    """

    def __init__(
        self,
        success_threshold: float = 0.95,
        crash_radius: float = 15,
        max_distance: float = 800,
        max_time: float = 12236,
    ) -> None:
        self.observation_space = gym.spaces.Box(
            np.concatenate(
                (
                    [-np.inf] * 3,  # position
                    [-np.inf] * 3,  # velocity
                    [0],  # sun angle
                    [0],  # num inspected
                    [-1] * 3,  # nearest cluster unit vector
                    [-1] * 3,  # priority vector unit vector
                    [0],  # weight inspected
                )
            ),
            np.concatenate(
                (
                    [np.inf] * 3,  # position
                    [np.inf] * 3,  # velocity
                    [2 * np.pi],  # sun angle
                    [100],  # num inspected
                    [1] * 3,  # nearest cluster unit vector
                    [1] * 3,  # priority vector unit vector
                    [1],  # weight inspected
                )
            ),
            shape=(15,),
        )

        self.action_space = gym.spaces.Box(-1, 1, shape=(3,))

        # Environment parameters
        self.crash_radius = crash_radius
        self.max_distance = max_distance
        self.max_time = max_time
        self.success_threshold = success_threshold

        # Episode level information
        self.prev_state = None
        self.prev_num_inspected = 0
        self.prev_weight_inspected = 0.0

    def reset(
        self, *, seed: int | None = None, options: dict[str, typing.Any] | None = None
    ) -> tuple[typing.Any, dict[str, typing.Any]]:
        super().reset(seed=seed, options=options)
        self._init_sim()  # sim is light enough we just reconstruct it
        self.simulator.reset()
        obs, info = self._get_obs(), self._get_info()
        self.prev_state = None
        self.prev_num_inspected = 0
        self.prev_weight_inspected = 0.0
        return obs, info

    def _init_sim(self):
        # Initialize spacecraft, sun, and simulator
        priority_vector = self.np_random.uniform(-1, 1, size=3)
        priority_vector /= np.linalg.norm(priority_vector)  # convert to unit vector
        self.chief = sim.Target(
            name="chief",
            num_points=100,
            radius=1,
            priority_vector=priority_vector,
        )
        self.deputy = sim.Inspector(
            name="deputy",
            position=utils.polar_to_cartesian(
                r=self.np_random.uniform(50, 100),
                theta=self.np_random.uniform(0, 2 * np.pi),
                phi=self.np_random.uniform(-np.pi / 2, np.pi / 2),
            ),
            velocity=utils.polar_to_cartesian(
                r=self.np_random.uniform(0, 0.3),
                theta=self.np_random.uniform(0, 2 * np.pi),
                phi=self.np_random.uniform(-np.pi / 2, np.pi / 2),
            ),
            fov=np.pi,
            focal_length=1,
        )
        self.sun = sim.Sun(theta=self.np_random.uniform(0, 2 * np.pi))
        self.simulator = sim.InspectionSimulator(
            frame_rate=0.1,
            inspectors=[self.deputy],
            targets=[self.chief],
            sun=self.sun,
        )

    def step(
        self, action: typing.Any
    ) -> tuple[typing.Any, typing.SupportsFloat, bool, bool, dict[str, typing.Any]]:
        # Store previous simulator state
        self.prev_state = self.sim_state.copy()
        self.prev_num_inspected = (
            self.chief.inspection_points.get_num_points_inspected()
        )
        self.prev_weight_inspected = (
            self.chief.inspection_points.get_total_weight_inspected()
        )

        # Update simulator state
        self.deputy.add_control(action)
        self.simulator.step()

        # Get info from simulator
        observation = self._get_obs()
        reward = self._get_reward()
        info = self._get_info()
        terminated = self._get_terminated()
        truncated = self.simulator.sim_time > self.max_time
        return observation, reward, terminated, truncated, info

    def _get_obs(self):
        obs = self.observation_space.sample()
        obs[:3] = self.deputy.position
        obs[3:6] = self.deputy.velocity
        obs[6] = self.sun.theta
        obs[7] = self.chief.inspection_points.get_num_points_inspected()
        obs[8:11] = self.chief.inspection_points.kmeans_find_nearest_cluster(
            camera=self.deputy.camera, sun=self.sun
        )
        obs[11:14] = self.chief.inspection_points.priority_vector
        obs[14] = self.chief.inspection_points.get_total_weight_inspected()
        return obs

    def _get_info(self):
        return {}

    def _get_reward(self):
        reward = 0

        # Dense rewards
        reward += r.weighted_observed_points_reward(
            chief=self.chief, prev_weight_inspected=self.prev_weight_inspected
        )
        reward += r.delta_v_reward(
            v=self.deputy.velocity, prev_v=self.prev_state["deputy"][3:6]
        )

        # Sparse rewards
        reward += (
            r.weighted_inspection_success_reward(
                chief=self.chief, total_weight=self.success_threshold
            )
            if utils.closest_fft_distance(chief=self.chief, deputy=self.deputy)
            < self.crash_radius
            else -1.0
        )
        reward += r.crash_reward(
            chief=self.chief, deputy=self.deputy, crash_radius=self.crash_radius
        )

        return reward

    def _get_terminated(self):
        # Get state info
        d = utils.rel_dist(pos1=self.chief.position, pos2=self.deputy.position)

        # Determine if in terminal state
        crash = d < self.crash_radius
        all_inspected = self.prev_weight_inspected >= self.success_threshold

        return crash or all_inspected

    def _get_truncated(self):
        d = utils.rel_dist(pos1=self.chief.position, pos2=self.deputy.position)
        oob = d > self.max_distance
        timeout = self.simulator.sim_time > self.max_time
        return oob or timeout

    @property
    def sim_state(self) -> dict:
        state = {
            "deputy": self.deputy.state,
            "chief": self.chief.state,
        }
        return state



"""
--------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core  Extension.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------

This module implements the Reward Functions and Reward Validators specific to the docking task.
"""
import math
from collections import OrderedDict

import numpy as np
from corl.libraries.state_dict import StateDict
from corl.rewards.reward_func_base import RewardFuncBase, RewardFuncBaseValidator
from corl.simulators.common_platform_utils import get_platform_by_name
from numpy_ringbuffer import RingBuffer

from safe_autonomy_sims.utils import get_relative_position, get_relative_velocity, max_vel_violation


class DockingTimeRewardValidator(RewardFuncBaseValidator):
    """
    A configuration validator for the DockingTimeRewardValidator Reward Function.

    Attributes
    ----------
    scale : float
        Scalar value to adjust magnitude of the reward.
    step_size : float
        The size of one simulation step (sec).
    """
    scale: float
    step_size: float


# Change name to TimeReward?
class DockingTimeReward(RewardFuncBase):
    """
    This reward function allocates reward based on the number of time steps (BaseSimulator.step() method calls) that
    have passed in a single episode.
    """

    def __init__(self, **kwargs):
        self.config: DockingTimeRewardValidator
        super().__init__(**kwargs)

    @staticmethod
    def get_validator():
        return DockingTimeRewardValidator

    def __call__(
        self,
        observation: OrderedDict,
        action,
        next_observation: OrderedDict,
        state: StateDict,
        next_state: StateDict,
        observation_space: StateDict,
        observation_units: StateDict,
    ) -> float:

        reward = self.config.scale * self.config.step_size

        return reward


class DockingDistanceChangeRewardValidator(RewardFuncBaseValidator):
    """
    A configuration validator for the DockingTimeRewardValidator Reward Function.

    Attributes
    ----------
    scale : float
        Scalar value to adjust magnitude of the reward.
    reference_position_sensor_name: str
        The name of the sensor responsible for returning the relative position of a reference entity.
    """
    scale: float
    reference_position_sensor_name: str = "reference_position"


class DockingDistanceChangeReward(RewardFuncBase):
    """
    This RewardFuncBase extension is responsible for calculating the reward associated with a change in agent position.

    """

    def __init__(self, **kwargs):
        self.config: DockingDistanceChangeRewardValidator
        super().__init__(**kwargs)
        self._dist_buffer = RingBuffer(capacity=2, dtype=float)

    @staticmethod
    def get_validator():
        return DockingDistanceChangeRewardValidator

    def __call__(
        self,
        observation: OrderedDict,
        action,
        next_observation: OrderedDict,
        state: StateDict,
        next_state: StateDict,
        observation_space: StateDict,
        observation_units: StateDict,
    ) -> float:

        # get relative dist
        # Assumes one platfrom per agent!
        platform = get_platform_by_name(next_state, self.config.platform_names[0])
        relative_position = get_relative_position(platform, self.config.reference_position_sensor_name)
        distance = np.linalg.norm(relative_position)

        self._dist_buffer.append(distance)

        reward = 0.0

        # TODO intialize distance buffer from initial state
        if len(self._dist_buffer) == 2:
            reward = self.config.scale * (self._dist_buffer[1] - self._dist_buffer[0])

        return reward


class DockingDistanceExponentialChangeRewardValidator(RewardFuncBaseValidator):
    """
    A configuration validator for the DockingDistanceExponentialChangeReward Reward Function.

    Attributes
    ----------
    c : float
        Scale factor of exponential distance function
    a : float
        Exponential coefficient of exponential distance function. Do not specify if `pivot` is specified.
    pivot : float
        Divisor of the exponential coefficient setting the scale of input values. Do not specify if `a` is specified.
    pivot_ratio : float
        logarithmic scale of exponential scaling coefficient. Do not specify if `a` is specified.
        determines portion of cumulative reward between distances <pivot and >pivot.
    scale : float
        Reward scaling value.
    reference_position_sensor_name: str
        The name of the sensor responsible for returning the relative position of a reference entity.
    """
    c: float = 2.0
    a: float = math.inf
    pivot: float = math.inf
    pivot_ratio: float = 2.0
    scale: float = 1.0
    reference_position_sensor_name: str = "reference_position"


class DockingDistanceExponentialChangeReward(RewardFuncBase):
    """
    Calculates an exponential reward based on the change in distance of the agent.
    Reward is based on the multiplicative scale factor to the exponential potential function:
        reward = psi(||r(t+1)||) - psi(||r(t)||)
        where psi(x) = ce^(-a*x)

    We choose a to be
        a = ln(pivot_ratio)/pivot
    This means that the reward accumulated going from distance r=pivot to r=0 is R=c(1 - 1/pivot_ratio)
        and the reward accumulated going from distance r=inf to r=pivot is c(1/pivot_ratio)

    Alternatively a can be set directly instead of using pivot and pivot ratio
    """

    def __init__(self, **kwargs):
        self.config: DockingDistanceExponentialChangeRewardValidator
        super().__init__(**kwargs)
        self._dist_buffer = RingBuffer(capacity=2, dtype=float)

        assert not (self.config.a == math.inf and self.config.pivot == math.inf), "Both 'a' and 'pivot' cannot be specified."
        assert self.config.a != math.inf or self.config.pivot != math.inf, "Either 'a' or 'pivot' must be specified."

        if self.config.a != math.inf:
            self.a = self.config.a
        else:
            self.a = math.log(self.config.pivot_ratio) / self.config.pivot

        self.c = self.config.c
        self.scale = self.config.scale

    @staticmethod
    def get_validator():
        return DockingDistanceExponentialChangeRewardValidator

    @property
    def prev_dist(self):
        """
        The previous distance of the agent from the target.

        Returns
        -------
        float
            The previous distance of the agent from the target.
        """
        return self._dist_buffer[0]

    @property
    def curr_dist(self):
        """
        The current distance of the agent from the target.

        Returns
        -------
        float
            The current distance of the agent from the target.
        """
        return self._dist_buffer[1]

    def update_dist(self, dist):
        """
        Store the current distance from the agent to the target.

        Parameters
        ----------
        dist: float
            The current distance of the agent from the target.
        """
        self._dist_buffer.append(dist)

    def __call__(
        self,
        observation: OrderedDict,
        action,
        next_observation: OrderedDict,
        state: StateDict,
        next_state: StateDict,
        observation_space: StateDict,
        observation_units: StateDict,
    ) -> float:

        # get relative dist
        # assumes one platform per agent!
        platform = get_platform_by_name(next_state, self.config.platform_names[0])
        relative_position = get_relative_position(platform, self.config.reference_position_sensor_name)
        distance = np.linalg.norm(relative_position)

        self.update_dist(distance)

        reward = 0.0

        # TODO initialize distance buffer from initial state
        if len(self._dist_buffer) == 2:
            reward = self.c * (math.exp(-self.a * self.curr_dist) - math.exp(-self.a * self.prev_dist))
            reward = self.scale * reward

        return reward


class DockingDeltaVRewardValidator(RewardFuncBaseValidator):
    """
    A configuration validator for the DockingDeltaVReward Reward Function.

    Attributes
    ----------
    scale : float
        A scalar value applied to reward.
    bias : float
        A bias value added to the reward.
    step_size : float
        Size of a single simulation step.
    mass : float
        The mass (kg) of the agent's spacecraft.
    """
    scale: float
    bias: float = 0.0
    step_size: float = 1.0
    mass: float


class DockingDeltaVReward(RewardFuncBase):
    """
    Calculates reward based on the agent's fuel consumption measured in delta-v.
    """

    def __init__(self, **kwargs):
        self.config: DockingDeltaVRewardValidator
        super().__init__(**kwargs)
        self.bias = self.config.bias
        self.scale = self.config.scale
        self.mass = self.config.mass
        self.step_size = self.config.step_size

    def delta_v(self, state):
        """
        Get change in agent's velocity from the current state.

        Parameters
        ----------
        state: StateDict
            The current state of the system.

        Returns
        -------
        d_v: float
            The agent's change in velocity
        """
        deputy = get_platform_by_name(state, self.config.platform_names[0])
        control_vec = deputy.get_applied_action()
        d_v = np.sum(np.abs(control_vec.m)) / self.mass * self.step_size
        return d_v

    @staticmethod
    def get_validator():
        return DockingDeltaVRewardValidator

    def __call__(
        self,
        observation: OrderedDict,
        action,
        next_observation: OrderedDict,
        state: StateDict,
        next_state: StateDict,
        observation_space: StateDict,
        observation_units: StateDict,
    ) -> float:

        reward = self.scale * self.delta_v(next_state) + self.bias
        return reward


class DockingVelocityConstraintRewardValidator(RewardFuncBaseValidator):
    """
    A configuration validator for the DockingVelocityConstraintReward Reward function.

    Attributes
    ----------
    scale : float
        Scalar value to adjust magnitude of the reward.
    bias : float
        Y intercept of the linear region of the velocity constraint function.
    step_size : float
        The size of one simulation step (sec).
    docking_region_radius : float
        The radius of the docking region in meters.
    velocity_threshold : float
        The maximum tolerated velocity within docking region without crashing.
    threshold_distance : float
        The distance at which the velocity constraint reaches a minimum (typically the docking region radius).
    slope : float
        The slope of the linear region of the velocity constraint function.
    mean_motion : float
        Orbital mean motion of Hill's reference frame's circular orbit in rad/s, by default 0.001027.
    lower_bound : bool
        If True, the function enforces a minimum velocity constraint on the agent's platform.
    reference_position_sensor_name: str
        The name of the sensor responsible for returning the relative position of a reference entity.
    reference_velocity_sensor_name: str
        The name of the sensor responsible for returning the relative velocity of a reference entity.
    """
    scale: float
    bias: float = 0.0
    step_size: float = 1.0
    velocity_threshold: float
    threshold_distance: float
    slope: float = 2.0
    mean_motion: float = 0.001027
    lower_bound: bool = False
    reference_position_sensor_name: str = "reference_position"
    reference_velocity_sensor_name: str = "reference_velocity"


class DockingVelocityConstraintReward(RewardFuncBase):
    """
    Calculates reward based on agent's violation of the velocity constraint.
    """

    def __init__(self, **kwargs):
        self.config: DockingVelocityConstraintRewardValidator
        super().__init__(**kwargs)
        self.bias = self.config.bias
        self.scale = self.config.scale
        self.step_size = self.config.step_size
        self.velocity_threshold = self.config.velocity_threshold
        self.threshold_distance = self.config.threshold_distance
        self.slope = self.config.slope
        self.mean_motion = self.config.mean_motion
        self.lower_bound = self.config.lower_bound

        self.agent_name = self.config.agent_name

    @staticmethod
    def get_validator():
        return DockingVelocityConstraintRewardValidator

    def __call__(
        self,
        observation: OrderedDict,
        action,
        next_observation: OrderedDict,
        state: StateDict,
        next_state: StateDict,
        observation_space: StateDict,
        observation_units: StateDict,
    ) -> float:

        # Get relative position and velocity
        # Assumes one platfrom per agent!
        platform = get_platform_by_name(next_state, self.config.platform_names[0])
        relative_position = get_relative_position(platform, self.config.reference_position_sensor_name)
        relative_velocity = get_relative_velocity(platform, self.config.reference_velocity_sensor_name)

        violated, violation = max_vel_violation(
            relative_position,
            relative_velocity,
            self.config.velocity_threshold,
            self.config.threshold_distance,
            self.config.mean_motion,
            self.config.lower_bound,
            slope=self.config.slope,
        )

        reward = 0.0

        if violated:
            reward = self.scale * violation + self.bias

        return reward


class DockingSuccessRewardValidator(RewardFuncBaseValidator):
    """
    A configuration validator for the DockingSuccessRewardValidator Reward Function.

    Attributes
    ----------
    scale : float
        Scalar value to adjust magnitude of the reward.
    timeout : float
        The max time for an episode.
    docking_region_radius : float
        The radius of the docking region in meters.
    velocity_threshold : float
        The maximum tolerated velocity within docking region without crashing.
    threshold_distance : float
        The distance at which the velocity constraint reaches a minimum (typically the docking region radius).
    slope : float
        The slope of the linear region of the velocity constraint function.
    mean_motion : float
        Orbital mean motion of Hill's reference frame's circular orbit in rad/s, by default 0.001027.
    lower_bound : bool
        If True, the function enforces a minimum velocity constraint on the agent's platform.
    reference_position_sensor_name: str
        The name of the sensor responsible for returning the relative position of a reference entity.
    reference_velocity_sensor_name: str
        The name of the sensor responsible for returning the relative velocity of a reference entity.
    """
    scale: float
    timeout: float
    docking_region_radius: float
    velocity_threshold: float
    threshold_distance: float
    slope: float = 2.0
    mean_motion: float = 0.001027
    lower_bound: bool = False
    reference_position_sensor_name: str = "reference_position"
    reference_velocity_sensor_name: str = "reference_velocity"


class DockingSuccessReward(RewardFuncBase):
    """
    This Reward Function is responsible for calculating the reward associated with a successful docking.
    """

    def __init__(self, **kwargs) -> None:
        self.config: DockingSuccessRewardValidator
        super().__init__(**kwargs)

    @staticmethod
    def get_validator():
        """
        Method to return class's Validator.
        """
        return DockingSuccessRewardValidator

    def __call__(
        self,
        observation: OrderedDict,
        action,
        next_observation: OrderedDict,
        state: StateDict,
        next_state: StateDict,
        observation_space: StateDict,
        observation_units: StateDict,
    ) -> float:

        platform = get_platform_by_name(next_state, self.config.platform_names[0])
        sim_time = platform.sim_time

        # Get relative position and velocity
        # Assumes one platfrom per agent!
        relative_position = get_relative_position(platform, self.config.reference_position_sensor_name)
        relative_velocity = get_relative_velocity(platform, self.config.reference_velocity_sensor_name)

        distance = np.linalg.norm(relative_position)

        in_docking = distance <= self.config.docking_region_radius

        violated, _ = max_vel_violation(
            relative_position,
            relative_velocity,
            self.config.velocity_threshold,
            self.config.threshold_distance,
            self.config.mean_motion,
            self.config.lower_bound,
            slope=self.config.slope,
        )

        reward = 0.0

        if in_docking and not violated:
            reward = self.config.scale
            if self.config.timeout:
                # Add time reward component, if timeout specified
                reward += 1 - (sim_time / self.config.timeout)

        return reward


class DockingFailureRewardValidator(RewardFuncBaseValidator):
    """
    A configuraion validator for the DockingFailureRewardValidator Reward Function.

    Attributes
    ----------
    timeout_reward : float
        Reward (penalty) associated with a failure by exceeding max episode time.
    distance_reward : float
        Reward (penalty) associated with a failure by exceeding max distance.
    crash_reward : float
        Reward (penalty) associated with a failure by crashing.
    timeout : float
        Max episode time.
    max_goal_distance : float
        Max distance from the goal.
    docking_region_radius : float
        Radius of the docking region in meters.
    velocity_threshold : float
        The maximum tolerated velocity within docking region without crashing.
    threshold_distance : float
        The distance at which the velocity constraint reaches a minimum (typically the docking region radius).
    slope : float
        The slope of the linear region of the velocity constraint function.
    mean_motion : float
        Orbital mean motion of Hill's reference frame's circular orbit in rad/s, by default 0.001027.
    lower_bound : bool
        If True, the function enforces a minimum velocity constraint on the agent's platform.
    reference_position_sensor_name: str
        The name of the sensor responsible for returning the relative position of a reference entity.
    reference_velocity_sensor_name: str
        The name of the sensor responsible for returning the relative velocity of a reference entity.
    """
    timeout_reward: float
    distance_reward: float
    crash_reward: float
    timeout: float
    max_goal_distance: float
    docking_region_radius: float
    velocity_threshold: float
    threshold_distance: float
    slope: float = 2.0
    mean_motion: float = 0.001027
    lower_bound: bool = False
    reference_position_sensor_name: str = "reference_position"
    reference_velocity_sensor_name: str = "reference_velocity"


class DockingFailureReward(RewardFuncBase):
    """
    This Reward Function is responsible for calculating the reward (penalty) associated with a failed episode.
    """

    def __init__(self, **kwargs) -> None:
        self.config: DockingFailureRewardValidator
        super().__init__(**kwargs)

    @staticmethod
    def get_validator():
        """
        Method to return class's Validator.
        """
        return DockingFailureRewardValidator

    def __call__(
        self,
        observation: OrderedDict,
        action,
        next_observation: OrderedDict,
        state: StateDict,
        next_state: StateDict,
        observation_space: StateDict,
        observation_units: StateDict,
    ) -> float:

        platform = get_platform_by_name(next_state, self.config.platform_names[0])
        sim_time = platform.sim_time

        # TODO: update to chief location when multiple platforms enabled
        # Get relative position and velocity
        # Assumes one platfrom per agent!
        relative_position = get_relative_position(platform, self.config.reference_position_sensor_name)
        relative_velocity = get_relative_velocity(platform, self.config.reference_velocity_sensor_name)
        distance = np.linalg.norm(relative_position)

        in_docking = distance <= self.config.docking_region_radius

        violated, _ = max_vel_violation(
            relative_position,
            relative_velocity,
            self.config.velocity_threshold,
            self.config.threshold_distance,
            self.config.mean_motion,
            self.config.lower_bound,
            slope=self.config.slope,
        )

        reward = 0.0

        if sim_time >= self.config.timeout:
            # episode reached max time
            reward = self.config.timeout_reward
        elif distance >= self.config.max_goal_distance:
            # agent exceeded max distance from goal
            reward = self.config.distance_reward
        elif in_docking and violated:
            # agent exceeded velocity constraint within docking region
            reward = self.config.crash_reward

        return reward



"""
--------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core  Extension.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------

This module implements a glue which returns the dot product between the deputy orientation unit
vector and the unit vector pointing from the deputy to the chief.

Author: Kochise Bennett
"""
import typing
from collections import OrderedDict
from functools import cached_property

import gymnasium
import numpy as np
from corl.glues.base_multi_wrapper import BaseMultiWrapperGlue, BaseMultiWrapperGlueValidator
from corl.libraries.property import BoxProp, DictProp
from corl.libraries.units import corl_get_ureg


class DotProductGlueValidator(BaseMultiWrapperGlueValidator):
    """
    A configuration validator for the DotProductGlue.

    Attributes
    ----------
    normalize_obs_vectors: bool
        Naormalize vectors before taking dot product
    """
    normalize_vectors: bool = False


class DotProductGlue(BaseMultiWrapperGlue):
    """
    Computes dot product between the deputy orientation unit vector and the
    unit vector pointing from the deputy to the chief
    """

    class Fields:
        """
        Fields in this glue
        """
        DIRECT_OBSERVATION = "direct_observation"

    @staticmethod
    def get_validator() -> typing.Type[DotProductGlueValidator]:
        return DotProductGlueValidator

    def get_unique_name(self) -> str:
        glue_name0 = self.glues()[0].get_unique_name()
        glue_name1 = self.glues()[1].get_unique_name()
        return glue_name0 + "_DotProduct_" + glue_name1

    @cached_property
    def observation_prop(self):
        prop = BoxProp(low=[-100], high=[100], unit="dimensionless")
        return DictProp(spaces={self.Fields.DIRECT_OBSERVATION: prop})

    @cached_property
    def normalized_observation_space(self) -> typing.Optional[gymnasium.spaces.Space]:
        """
        passthrough property
        """
        return self.observation_space

    @cached_property
    def observation_space(self):
        d = gymnasium.spaces.dict.Dict()
        d.spaces[self.Fields.DIRECT_OBSERVATION] = gymnasium.spaces.Box(-1.0, 1.0, shape=(1, ), dtype=np.float32)
        return d

    def get_observation(self, other_obs: OrderedDict, obs_space: OrderedDict, obs_units: OrderedDict):
        glue0 = self.glues()[0]
        glue1 = self.glues()[1]

        obs0 = glue0.get_observation(other_obs, obs_space, obs_units)[glue0.Fields.DIRECT_OBSERVATION].m
        obs1 = glue1.get_observation(other_obs, obs_space, obs_units)[glue1.Fields.DIRECT_OBSERVATION].m

        dot_product = np.dot(obs0, obs1)

        if self.config.normalize_vectors:
            dot_product = dot_product / (np.linalg.norm(obs0) * np.linalg.norm(obs1) + 1e-5)

            dot_product = np.clip(dot_product, -1.0, 1.0)

        d = OrderedDict()
        d[self.Fields.DIRECT_OBSERVATION] = corl_get_ureg().Quantity(np.array([dot_product], dtype=np.float32), "dimensionless")
        return d



"""
--------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core  Extension.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------

This module implements initializers for CWH platforms.
"""

import typing

import numpy as np
from corl.libraries.units import Quantity
from scipy.spatial.transform import Rotation

from safe_autonomy_sims.simulators.initializers.initializer import BaseInitializerWithPint, InitializerValidator, strip_units_from_dict
from safe_autonomy_sims.utils import velocity_limit


class CWH3DRadialInitializer(BaseInitializerWithPint):
    """
    This class handles the initialization of agent reset parameters for the cwh 3D environment.
    Both position and velocity are initialized radially (with radius and angles) to allow for control over magnitude and direction
    of the resulting vectors.
    """

    param_units = {
        'radius': 'meters',
        'azimuth_angle': 'radians',
        'elevation_angle': 'radians',
        'vel_mag': 'meter/second',
        'vel_azimuth_angle': 'radians',
        'vel_elevation_angle': 'radians',
    }

    def compute(self, **kwargs):
        return self._compute_with_units(**strip_units_from_dict(kwargs))

    # TODO: change name or function to reflect unit use. Currently, no units used.
    def _compute_with_units(
        self,
        radius: float,
        azimuth_angle: float,
        elevation_angle: float,
        vel_mag: float,
        vel_azimuth_angle: float,
        vel_elevation_angle: float,
    ) -> typing.Dict:
        """Computes radial initial conditions for cwh 3d

        Parameters
        ----------
        radius : float
            radius from origin. meters
        azimuth_angle : float
            location azimuthal angle in spherical coordinates (right hand convention). rad
        elevation_angle : float
            location elevation angle from x-y plane. Positive angles = positive z. rad
        vel_mag : float
            magnitude of velocity vector. meter/second
        vel_azimuth_angle : float
            velocity vector azimuthal angle in spherical coordinates (right hand convention). rad
        vel_elevation_angle : float
            velocity vector elevation angle from x-y plane. Positive angles = positive z. rad

        Returns
        -------
        typing.Dict
            initial conditions of platform
        """

        x = radius * np.cos(azimuth_angle) * np.cos(elevation_angle)
        y = radius * np.sin(azimuth_angle) * np.cos(elevation_angle)
        z = radius * np.sin(elevation_angle)

        x_dot = vel_mag * np.cos(vel_azimuth_angle) * np.cos(vel_elevation_angle)
        y_dot = vel_mag * np.sin(vel_azimuth_angle) * np.cos(vel_elevation_angle)
        z_dot = vel_mag * np.sin(vel_elevation_angle)

        return {
            'x': x,
            'y': y,
            'z': z,
            'x_dot': x_dot,
            'y_dot': y_dot,
            'z_dot': z_dot,
        }


class CWH3DENMTInitializerValidator(InitializerValidator):
    """
    A configuration validator for CWH3DENMTInitializer.

    Attributes
    ----------
    mean_motion : float
        Orbital mean motion of Hill's reference frame's circular orbit in rad/s
    """

    mean_motion: float = 0.001027


class CWH3DENMTInitializer(BaseInitializerWithPint):
    """
    This class handles the initialization of agent reset parameters for the cwh 3D environment.
    It ensures that each agent starts on an elliptical Natural Motion Trajectory (eNMT)
    """

    param_units = {
        'radius': 'meters',
        'azimuth_angle': 'radians',
        'elevation_angle': 'radians',
        'z_dot': 'meter/second',
    }

    @staticmethod
    def get_validator():
        """
        Returns
        -------
        CWH3DENMTInitializerValidator
            Config validator for the CWH3DENMTInitializerValidator.
        """
        return CWH3DENMTInitializerValidator

    def compute(self, **kwargs):
        return self._compute_with_units(**strip_units_from_dict(kwargs))

    def _compute_with_units(
        self,
        radius: float,
        azimuth_angle: float,
        elevation_angle: float,
        z_dot: float,
    ) -> typing.Dict:
        """Computes eNMT initial conditions for cwh 3d

        Parameters
        ----------
        radius : float
            radius from origin. meters
        azimuth_angle : float
            location azimuthal angle in spherical coordinates (right hand convention). rad
        elevation_angle : float
            location elevation angle from x-y plane. Positive angles = positive z. rad
        z_dot : float
            The reset value for the z velocity of the agent

        Returns
        -------
        typing.Dict
            initial conditions of platform
        """
        x = radius * np.cos(azimuth_angle) * np.cos(elevation_angle)
        y = radius * np.sin(azimuth_angle) * np.cos(elevation_angle)
        z = radius * np.sin(elevation_angle)

        x_dot = self.config.mean_motion * y / 2
        y_dot = -2 * self.config.mean_motion * x

        return {
            'x': x,
            'y': y,
            'z': z,
            'x_dot': x_dot,
            'y_dot': y_dot,
            'z_dot': z_dot,
        }


class Docking3DRadialInitializerValidator(InitializerValidator):
    """
    A configuration validator for Docking3DInitializer.

    Attributes
    ----------
    velocity_threshold : float
        The maximum tolerated velocity within docking region without crashing.
    threshold_distance : float
        The distance at which the velocity constraint reaches a minimum (typically the docking region radius).
    slope : float
        The slope of the linear region of the velocity constraint function.
    mean_motion : float
        Orbital mean motion of Hill's reference frame's circular orbit in rad/s
    """

    velocity_threshold: float
    threshold_distance: float
    slope: float = 2.0
    mean_motion: float = 0.001027


class Docking3DRadialInitializer(BaseInitializerWithPint):
    """
    This class handles the initialization of agent reset parameters for the cwh 3D docking environment.
    It ensures that the initial velocity of the deputy does not violate the maximum velocity safety constraint.
    """

    param_units = {
        'radius': 'meters',
        'azimuth_angle': 'radians',
        'elevation_angle': 'radians',
        'vel_max_ratio': None,
        'vel_azimuth_angle': 'radians',
        'vel_elevation_angle': 'radians',
    }

    def __init__(self, config):
        super().__init__(config)
        self.config: Docking3DRadialInitializerValidator

    @staticmethod
    def get_validator():
        """
        Returns
        -------
        Docking3DInitializerValidator
            Config validator for the Docking3DInitializerValidator.
        """
        return Docking3DRadialInitializerValidator

    def compute(self, **kwargs):
        return self._compute_with_units(**strip_units_from_dict(kwargs))

    def _compute_with_units(
        self,
        radius: float,
        azimuth_angle: float,
        elevation_angle: float,
        vel_max_ratio: float,
        vel_azimuth_angle: float,
        vel_elevation_angle: float,
    ) -> typing.Dict:
        """Computes radial initial conditions for 3d docking problem

        Parameters
        ----------
        radius : float
            radius from origin. meters
        azimuth_angle : float
            location azimuthal angle in spherical coordinates (right hand convention). rad
        elevation_angle : float
            location elevation angle from x-y plane. Positive angles = positive z. rad
        vel_max_ratio : float
            Ratio of max safe velocity to assign to initial velocity magnitde
        vel_azimuth_angle : float
            velocity vector azimuthal angle in spherical coordinates (right hand convention). rad
        vel_elevation_angle : float
            velocity vector elevation angle from x-y plane. Positive angles = positive z. rad

        Returns
        -------
        typing.Dict
            initial conditions of platform
        """
        # contend with corl Quantities
        radius = radius.value if isinstance(radius, Quantity) else radius
        azimuth_angle = azimuth_angle.value if isinstance(azimuth_angle, Quantity) else azimuth_angle
        elevation_angle = elevation_angle.value if isinstance(elevation_angle, Quantity) else elevation_angle
        vel_max_ratio = vel_max_ratio.value if isinstance(vel_max_ratio, Quantity) else vel_max_ratio
        vel_azimuth_angle = vel_azimuth_angle.value if isinstance(vel_azimuth_angle, Quantity) else vel_azimuth_angle
        vel_elevation_angle = vel_elevation_angle.value if isinstance(vel_elevation_angle, Quantity) else vel_elevation_angle

        x = radius * np.cos(azimuth_angle) * np.cos(elevation_angle)
        y = radius * np.sin(azimuth_angle) * np.cos(elevation_angle)
        z = radius * np.sin(elevation_angle)

        distance = np.linalg.norm([x, y, z])

        vel_limit = velocity_limit(
            distance, self.config.velocity_threshold, self.config.threshold_distance, self.config.mean_motion, self.config.slope
        )

        vel_mag = vel_max_ratio * vel_limit

        x_dot = vel_mag * np.cos(vel_azimuth_angle) * np.cos(vel_elevation_angle)
        y_dot = vel_mag * np.sin(vel_azimuth_angle) * np.cos(vel_elevation_angle)
        z_dot = vel_mag * np.sin(vel_elevation_angle)

        return {
            'x': x,
            'y': y,
            'z': z,
            'x_dot': x_dot,
            'y_dot': y_dot,
            'z_dot': z_dot,
        }


class CWHSixDOFRadialInitializer(BaseInitializerWithPint):
    """
    This class handles the initialization of agent reset parameters for the cwh 3D environment.
    Both position and velocity are initialized radially (with radius and angles) to allow for control over magnitude and direction
        of the resulting vectors.
    """

    param_units = {
        'radius': 'meters',
        'azimuth_angle': 'radians',
        'elevation_angle': 'radians',
        'vel_mag': 'meter/second',
        'vel_azimuth_angle': 'radians',
        'vel_elevation_angle': 'radians',
        'wx': 'radian/second',
        'wy': 'radian/second',
        'wz': 'radian/second',
    }

    def compute(self, **kwargs):
        return self._compute_with_units(**strip_units_from_dict(kwargs))

    def _compute_with_units(
        self,
        radius: float,
        azimuth_angle: float,
        elevation_angle: float,
        vel_mag: float,
        vel_azimuth_angle: float,
        vel_elevation_angle: float,
        wx: float,
        wy: float,
        wz: float,
    ) -> typing.Dict:
        """Computes radial initial conditions for cwh 3d

        Parameters
        ----------
        radius : float
            radius from origin. meters
        azimuth_angle : float
            location azimuthal angle in spherical coordinates (right hand convention). rad
        elevation_angle : float
            location elevation angle from x-y plane. Positive angles = positive z. rad
        vel_mag : float
            magnitude of velocity vector. meter/second
        vel_azimuth_angle : float
            velocity vector azimuthal angle in spherical coordinates (right hand convention). rad
        vel_elevation_angle : float
            velocity vector elevation angle from x-y plane. Positive angles = positive z. rad

        Returns
        -------
        typing.Dict
            initial conditions of platform
        """

        x = radius * np.cos(azimuth_angle) * np.cos(elevation_angle)
        y = radius * np.sin(azimuth_angle) * np.cos(elevation_angle)
        z = radius * np.sin(elevation_angle)

        x_dot = vel_mag * np.cos(vel_azimuth_angle) * np.cos(vel_elevation_angle)
        y_dot = vel_mag * np.sin(vel_azimuth_angle) * np.cos(vel_elevation_angle)
        z_dot = vel_mag * np.sin(vel_elevation_angle)

        q = Rotation.random().as_quat()

        return {
            'x': x,
            'y': y,
            'z': z,
            'x_dot': x_dot,
            'y_dot': y_dot,
            'z_dot': z_dot,
            'q1': q[0],
            'q2': q[1],
            'q3': q[2],
            'q4': q[3],
            'wx': wx,
            'wy': wy,
            'wz': wz,
        }


class CWH3DRadialWithSunInitializer(BaseInitializerWithPint):
    """
    This class handles the initialization of agent reset parameters for the cwh 3D environment.
    Both position and velocity are initialized radially (with radius and angles) to allow for
    control over magnitude and direction of the resulting vectors.
    The sun angle is also passed through.
    """

    param_units = {
        'radius': 'meters',
        'azimuth_angle': 'radians',
        'elevation_angle': 'radians',
        'vel_mag': 'meter/second',
        'vel_azimuth_angle': 'radians',
        'vel_elevation_angle': 'radians',
        'sun_angle': 'radians',
    }

    def compute(self, **kwargs):
        return self._compute_with_units(**strip_units_from_dict(kwargs))

    def _compute_with_units(
        self,
        radius: float,
        azimuth_angle: float,
        elevation_angle: float,
        vel_mag: float,
        vel_azimuth_angle: float,
        vel_elevation_angle: float,
        sun_angle: float,
    ) -> typing.Dict:
        """Computes radial initial conditions for cwh 3d

        Parameters
        ----------
        radius : float
            radius from origin. meters
        azimuth_angle : float
            location azimuthal angle in spherical coordinates (right hand convention). rad
        elevation_angle : float
            location elevation angle from x-y plane. Positive angles = positive z. rad
        vel_mag : float
            magnitude of velocity vector. meter/second
        vel_azimuth_angle : float
            velocity vector azimuthal angle in spherical coordinates (right hand convention). rad
        vel_elevation_angle : float
            velocity vector elevation angle from x-y plane. Positive angles = positive z. rad
        sun_angle : float
            Initial angle of the sun. rad

        Returns
        -------
        typing.Dict
            initial conditions of platform
        """

        x = radius * np.cos(azimuth_angle) * np.cos(elevation_angle)
        y = radius * np.sin(azimuth_angle) * np.cos(elevation_angle)
        z = radius * np.sin(elevation_angle)

        x_dot = vel_mag * np.cos(vel_azimuth_angle) * np.cos(vel_elevation_angle)
        y_dot = vel_mag * np.sin(vel_azimuth_angle) * np.cos(vel_elevation_angle)
        z_dot = vel_mag * np.sin(vel_elevation_angle)

        return {
            'x': x,
            'y': y,
            'z': z,
            'x_dot': x_dot,
            'y_dot': y_dot,
            'z_dot': z_dot,
            'sun_angle': sun_angle,
        }



"""
--------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core  Extension.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------

This module implements common sensors for the safe autonomy environments.
"""
import abc
import typing

import numpy as np
from corl.simulators.base_parts import BaseSensor


class TransformSensor(BaseSensor):
    """
    Sensor part which applies a transformation to the measurement.
    """

    def _calculate_measurement(self, state: typing.Tuple) -> typing.Union[np.ndarray, typing.Tuple, typing.Dict]:
        measurement = self._raw_measurement(state)
        measurement = self._transform(measurement, state)
        return measurement

    def _raw_measurement(self, state):
        """
        Calculate the raw sensor measurement value.

        Parameters
        ----------
        state: typing.Tuple
            The current state of the environment used to obtain the measurement

        Returns
        -------
        typing.Union[np.ndarray, typing.Tuple, typing.Dict]
            The raw measurements from this sensor.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def _transform(self, measurement, state):
        """
        Perform a transformation on a measurement.

        Parameters
        ----------
        measurement: typing.Union[np.ndarray, typing.Tuple, typing.Dict]
            The measurement to be transformed.
        state: typing.Tuple
            The current state of the environment used to obtain the measurement.

        Returns
        -------
        typing.Union[np.ndarray, typing.Tuple, typing.Dict]
            The transformed measurements from this sensor.
        """
        raise NotImplementedError



"""
--------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core  Extension.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------

This module implements functions that define the terminal conditions
for the inspection environment.
"""
import typing
from collections import OrderedDict

import numpy as np
from corl.dones.done_func_base import DoneFuncBase, DoneFuncBaseValidator, DoneStatusCodes, SharedDoneFuncBase, SharedDoneFuncBaseValidator
from corl.libraries.environment_dict import DoneDict
from corl.libraries.state_dict import StateDict

from safe_autonomy_sims.utils import get_closest_fft_distance


class SuccessfulInspectionDoneValidator(DoneFuncBaseValidator):
    """
    A configuration validator for the SuccessfulInspectionDoneValidator.

    Attributes
    ----------
    inspection_entity_name: str
        The name of the entity under inspection.
    weight_threshold : float
        Points score value indicating success.
        By default None, so success occurs when all points are inspected
    """
    inspection_entity_name: str = "chief"
    weight_threshold: typing.Union[float, None] = None


class SuccessfulInspectionDoneFunction(DoneFuncBase):
    """
    A done function that determines if the deputy has successfully
    inspected the chief.

    Attributes
    ----------
    config: SuccessfulInspectionDoneValidator
        The function's validated configuration parameters
    """

    def __init__(self, **kwargs) -> None:
        self.config: SuccessfulInspectionDoneValidator
        super().__init__(**kwargs)

    @staticmethod
    def get_validator():
        """
        Returns configuration validator object for this done
        function.

        Returns
        -------
        SuccessfulInspectionDoneValidator
            Config validator for the SuccessfulInspectionDoneFunction.
        """
        return SuccessfulInspectionDoneValidator

    def __call__(
        self,
        observation: OrderedDict,
        action: OrderedDict,
        next_observation: OrderedDict,
        next_state: StateDict,
        observation_space: StateDict,
        observation_units: StateDict,
    ) -> bool:
        """
        Parameters
        ----------
        observation : OrderedDict
            the current observation
        action : OrderedDict
            the current action
        next_observation : OrderedDict
            the incoming observation
        next_state : StateDict
            the incoming state
        observation_space: StateDict
            the observation space
        observation_units: StateDict
            the observation units

        Returns
        -------
        done : bool
            Dictionary containing the done condition for the current agent.
        """

        if self.config.weight_threshold is not None:
            weight = next_state.inspection_points_map[self.config.inspection_entity_name].get_total_weight_inspected()
            done_check = weight >= self.config.weight_threshold
        else:
            inspection_points = next_state.inspection_points_map[self.config.inspection_entity_name]
            done_check = all(inspection_points.points_inspected_dict.values())

        done = bool(done_check)
        if done:
            next_state.episode_state[self.config.platform_name][self.name] = DoneStatusCodes.WIN

        return done


class SafeSuccessfulInspectionDoneValidator(SuccessfulInspectionDoneValidator):
    """
    A configuration validator for the SafeSuccessfulInspectionDoneFunction.

    Attributes
    ----------
    mean_motion : float
        orbital mean motion in rad/s of current Hill's reference frame
    crash_region_radius : float
        The radius of the crashing region in meters.
    fft_time_step : float
        Time step to compute the FFT trajectory. FFT is computed for 1 orbit.
    """
    mean_motion: float
    crash_region_radius: float
    fft_time_step: float = 1


class SafeSuccessfulInspectionDoneFunction(SuccessfulInspectionDoneFunction):
    """
    A done function that determines if the deputy has successfully
    inspected the chief.

    Considers if a Free Flight Trajectory once the episode ends
    **would not** result in a collision.

    Attributes
    ----------
    config: SafeSuccessfulInspectionDoneValidator
        The function's validated configuration parameters
    """

    def __init__(self, **kwargs) -> None:
        self.config: SafeSuccessfulInspectionDoneValidator
        super().__init__(**kwargs)

    @staticmethod
    def get_validator():
        """
        Returns configuration validator object for this done
        function.

        Returns
        -------
        SafeSuccessfulInspectionDoneValidator
            Config validator for the SafeSuccessfulInspectionDoneFunction."""
        return SafeSuccessfulInspectionDoneValidator

    def __call__(
        self,
        observation: OrderedDict,
        action: OrderedDict,
        next_observation: OrderedDict,
        next_state: StateDict,
        observation_space: StateDict,
        observation_units: StateDict,
    ) -> bool:
        """
        Parameters
        ----------
        observation : OrderedDict
            the current observation
        action : OrderedDict
            the current action
        next_observation : OrderedDict
            the incoming observation
        next_state : StateDict
            the incoming state
        observation_space: StateDict
            the observation space
        observation_units: StateDict
            the observation units

        Returns
        -------
        done : bool
            Dictionary containing the done condition for the current agent.
        """

        done = super().__call__(observation, action, next_observation, next_state, observation_space, observation_units)

        if done:
            pos = next_state.sim_platforms[self.config.platform_name].position
            vel = next_state.sim_platforms[self.config.platform_name].velocity
            state = np.concatenate((pos, vel))
            n = self.config.mean_motion
            times = np.arange(0, 2 * np.pi / n, self.config.fft_time_step)
            dist = get_closest_fft_distance(state, self.config.mean_motion, times)
            if dist >= self.config.crash_region_radius:
                next_state.episode_state[self.config.platform_name][self.name] = DoneStatusCodes.WIN
            else:
                # TODO: why is done set to False if deputy is within crash radius? Would crash not end episode?
                done = False

        return done


class CrashAfterSuccessfulInspectionDoneFunction(SuccessfulInspectionDoneFunction):
    """
    A done function that determines if the deputy has successfully
    inspected the chief.

    Considers if a Free Flight Trajectory once the episode ends
    **would** result in a collision.

    Attributes
    ----------
    config: SafeSuccessfulInspectionDoneValidator
        The function's validated configuration parameters
    """

    def __init__(self, **kwargs) -> None:
        self.config: SafeSuccessfulInspectionDoneValidator
        super().__init__(**kwargs)

    @staticmethod
    def get_validator():
        """
        Returns configuration validator object for this done
        function.

        Returns
        -------
        SafeSuccessfulInspectionDoneValidator
            Config validator for the CrashAfterSuccessfulInspectionDoneFunction.
        """
        return SafeSuccessfulInspectionDoneValidator

    def __call__(
        self,
        observation: OrderedDict,
        action: OrderedDict,
        next_observation: OrderedDict,
        next_state: StateDict,
        observation_space: StateDict,
        observation_units: StateDict,
    ) -> bool:
        """
        Parameters
        ----------
        observation : OrderedDict
            the current observation
        action : OrderedDict
            the current action
        next_observation : OrderedDict
            the incoming observation
        next_state : StateDict
            the incoming state
        observation_space: StateDict
            the observation space
        observation_units: StateDict
            the observation units

        Returns
        -------
        done : bool
            Dictionary containing the done condition for the current agent.
        """

        done = super().__call__(observation, action, next_observation, next_state, observation_space, observation_units)

        if done:
            pos = next_state.sim_platforms[self.config.platform_name].position
            vel = next_state.sim_platforms[self.config.platform_name].velocity
            state = np.concatenate((pos, vel))
            n = self.config.mean_motion
            times = np.arange(0, 2 * np.pi / n, self.config.fft_time_step)
            dist = get_closest_fft_distance(state, self.config.mean_motion, times)
            if dist < self.config.crash_region_radius:
                next_state.episode_state[self.config.platform_name][self.name] = DoneStatusCodes.LOSE
            else:
                done = False

        return done


class MultiagentSuccessfulInspectionDoneFunctionValidator(SharedDoneFuncBaseValidator):
    """
    A configuration validator for the MultiagentSuccessfulDockingDoneFunction.

    Attributes
    ----------
    inspection_entity_name: str
        The name of the entity under inspection.
    weight_threshold : float
        Points score value indicating success.
        By default None, so success occurs when all points are inspected
    """
    inspection_entity_name: str = "chief"
    weight_threshold: typing.Union[float, None] = None


class MultiagentSuccessfulInspectionDoneFunction(SharedDoneFuncBase):
    """
    A done function which determines whether every agent in the
    environment has reached a specified successful done condition.

    Attributes
    ----------
    config: MultiagentSuccessfulInspectionDoneFunctionValidator
        The function's validated configuration parameters
    """

    @staticmethod
    def get_validator() -> typing.Type[SharedDoneFuncBaseValidator]:
        """
        Returns the validator for this done function.

        Returns
        -------
        MultiagentSuccessfulDockingDoneFunctionValidator
            Config validator for the MultiagentSuccessfulInspectionDoneFunction.
        """
        return MultiagentSuccessfulInspectionDoneFunctionValidator

    def __call__(
        self,
        observation: OrderedDict,
        action: OrderedDict,
        next_observation: OrderedDict,
        next_state: StateDict,
        observation_space: StateDict,
        observation_units: StateDict,
        local_dones: DoneDict,
        local_done_info: OrderedDict
    ) -> DoneDict:
        """
        Parameters
        ----------
        observation : OrderedDict
            the current observation
        action : OrderedDict
            the current action
        next_observation : OrderedDict
            the incoming observation
        next_state : StateDict
            the incoming state
        observation_space: StateDict
            the observation space
        observation_units: StateDict
            the observation units

        Returns
        -------
        done : bool
            Dictionary containing the done condition for the current agent.
        """

        done = DoneDict()

        if self.config.weight_threshold is not None:
            weight = next_state.inspection_points_map[self.config.inspection_entity_name].get_total_weight_inspected()
            done_check = weight >= self.config.weight_threshold
        else:
            inspection_points = next_state.inspection_points_map[self.config.inspection_entity_name]
            done_check = all(inspection_points.points_inspected_dict.values())

        if done_check:
            for k in local_dones.keys():
                done[k] = True

        return done



from safe_autonomy_sims.pettingzoo.docking.multidocking_v0 import MultiDockingEnv
from safe_autonomy_sims.pettingzoo.inspection.multi_inspection_v0 import (
    MultiInspectionEnv,
)
from safe_autonomy_sims.pettingzoo.inspection.weighted_multi_inspection_v0 import (
    WeightedMultiInspectionEnv,
)
from safe_autonomy_sims.pettingzoo.inspection.sixdof_multi_inspection_v0 import (
    WeightedSixDofMultiInspectionEnv,
)


__all__ = [
    "MultiDockingEnv",
    "MultiInspectionEnv",
    "WeightedMultiInspectionEnv",
    "WeightedSixDofMultiInspectionEnv",
]



"""Utility functions for the docking environment"""

import numpy as np


def polar_to_cartesian(r: float, theta: float, phi: float) -> np.ndarray:
    """Convert polar coordinates to cartesian coordinates.

    Parameters
    ----------
    r : float
        radial distance
    theta : float
        azimuthal angle
    phi : float
        polar angle

    Returns
    -------
    np.ndarray
        cartesian coordinates
    """
    x = r * np.sin(phi) * np.cos(theta)
    y = r * np.sin(phi) * np.sin(theta)
    z = r * np.cos(phi)
    return np.array([x, y, z])


def rel_dist(pos1: np.ndarray, pos2: np.ndarray) -> float:
    """The relative distance between two positions.

    Parameters
    ----------
    pos1 : np.ndarray
        the first position
    pos2 : np.ndarray
        the second position

    Returns
    -------
    float
        the euclidean distance between the two positions
    """
    rel_d = np.linalg.norm(pos1 - pos2)
    return rel_d


def rel_vel(vel1: np.ndarray, vel2: np.ndarray) -> float:
    """The relative velocity between two velocities.

    Parameters
    ----------
    vel1 : np.ndarray
        the first velocity
    vel2 : np.ndarray
        the second velocity

    Returns
    -------
    float
        the relative velocity between the two velocities
    """
    rel_v = np.linalg.norm(vel1 - vel2)
    return rel_v


def delta_v(v: np.ndarray, prev_v: np.ndarray) -> np.ndarray:
    """The change in velocity

    Parameters
    ----------
    v : np.ndarray
        the current velocity
    prev_v : np.ndarray
        the previous velocity

    Returns
    -------
    float
        the change in velocity
    """
    v_norm = np.linalg.norm(v)
    prev_v_norm = np.linalg.norm(prev_v)
    return v_norm - prev_v_norm


def v_limit(
    chief_pos: np.ndarray,
    deputy_pos: np.ndarray,
    a: float = 2.0,
    n: float = 0.001027,
    v_max: float = 0.2,
    docking_radius: float = 0.5,
):
    """A linear velocity limit based on the distance between the chief
    and the deputy.

    $v_{limit} = v_{max} + an(d-r)$

    where:
    * $v_max$ is a maximum allowable velocity
    * $a$ is the slope of the linear velocity limit
    * $d$ is the distance between the chief and the deputy
    * $r$ is the radius of the docking region around the chief

    Parameters
    ----------
    chief_pos : np.ndarray
        the position of the chief
    deputy_pos : np.ndarray
        the position of the deputy
    a : float, optional
        slope of the linear velocity limit, by default 2.0
    n : float, optional
        deputy mean motion, by default 0.001027
    v_max : float, optional
        maximum allowable velocity, by default 0.2
    docking_radius : float, optional
        the radius of the docking region around the chief, by default 0.5

    Returns
    -------
    float
        velocity limit
    """
    v_limit = v_max + (
        a * n * (rel_dist(pos1=chief_pos, pos2=deputy_pos) - docking_radius)
    )
    return v_limit



"""
--------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core  Extension.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------

This module defines and registers the available CWH platform types to the proper simulators.
"""

from __future__ import annotations

from corl.libraries.plugin_library import PluginLibrary
from corl.simulators.base_platform_type import BasePlatformType

from safe_autonomy_sims.simulators.cwh_simulator import CWHSimulator
from safe_autonomy_sims.simulators.inspection_simulator import InspectionSimulator


class CWHAvailablePlatformTypes(BasePlatformType):
    """
    """

    @classmethod
    def match_model(cls, config: dict) -> bool:
        if "name" not in config:
            raise RuntimeError("Attempting to parse a PlatformType from name/model config, but both are not given!")

        if config["name"] == "CWH":
            return True
        if config["name"] == "CWHSixDOF":
            return True
        return False


PluginLibrary.add_platform_to_sim(CWHAvailablePlatformTypes, CWHSimulator)
PluginLibrary.add_platform_to_sim(CWHAvailablePlatformTypes, InspectionSimulator)



"""Utility functions for the docking environment"""

import numpy as np


def polar_to_cartesian(r: float, theta: float, phi: float) -> np.ndarray:
    """Convert polar coordinates to cartesian coordinates.

    Parameters
    ----------
    r : float
        radial distance
    theta : float
        azimuthal angle
    phi : float
        polar angle

    Returns
    -------
    np.ndarray
        cartesian coordinates
    """
    x = r * np.sin(phi) * np.cos(theta)
    y = r * np.sin(phi) * np.sin(theta)
    z = r * np.cos(phi)
    return np.array([x, y, z])


def rel_dist(state: dict):
    """The relative distance between the chief and the deputy.

    Parameters
    ----------
    state : dict
        the current state of the system

    Returns
    -------
    float
        the euclidean distance between the chief and the deputy
    """
    chief_pos = state["chief"][:3]
    deputy_pos = state["deputy"][:3]
    rel_d = np.linalg.norm(chief_pos - deputy_pos)
    return rel_d


def rel_vel(state: dict):
    """The relative velocity between the chief and the deputy.

    Parameters
    ----------
    state : dict
        the current state of the system

    Returns
    -------
    float
        the relative velocity between the chief and the deputy
    """
    chief_v = state["chief"][3:6]
    deputy_v = state["deputy"][3:6]
    rel_v = np.linalg.norm(chief_v - deputy_v)
    return rel_v


def delta_v(state: dict, prev_state: dict):
    """The change in velocity of the deputy.

    Parameters
    ----------
    state : dict
        the current simulation state
    prev_state : dict
        the previous simulation state

    Returns
    -------
    float
        the deputy's change in velocity
    """
    v = np.linalg.norm(state["deputy"][3:6])
    prev_v = np.linalg.norm(prev_state["deputy"][3:6])
    return v - prev_v


def v_limit(
    state: dict,
    a: float = 2.0,
    n: float = 0.001027,
    v_max: float = 0.2,
    docking_radius: float = 0.5,
):
    """A linear velocity limit based on the distance between the chief
    and the deputy.

    $v_{limit} = v_{max} + an(d-r)$

    where:
    * $v_max$ is a maximum allowable velocity
    * $a$ is the slope of the linear velocity limit
    * $d$ is the distance between the chief and the deputy
    * $r$ is the radius of the docking region around the chief

    Parameters
    ----------
    state : dict
        the current simulation state
    a : float, optional
        slope of the linear velocity limit, by default 2.0
    n : float, optional
        deputy mean motion, by default 0.001027
    v_max : float, optional
        maximum allowable velocity, by default 0.2
    docking_radius : float, optional
        the radius of the docking region around the chief, by default 0.5

    Returns
    -------
    float
        velocity limit
    """
    v_limit = v_max + (a * n * (rel_dist(state=state) - docking_radius))
    return v_limit



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------

This module implements general observation and action metrics
"""
import typing

import numpy as np
from corl.evaluation.episode_artifact import EpisodeArtifact
from corl.evaluation.metrics.generator import MetricGeneratorTerminalEventScope
from corl.evaluation.metrics.metric import Metric
from corl.evaluation.metrics.types.nonterminals.dict import Dict
from corl.evaluation.metrics.types.nonterminals.vector import Vector
from corl.evaluation.metrics.types.terminals.real import Real


class ObservationVector(MetricGeneratorTerminalEventScope):
    """
    Generates vector of Dicts for the observations calculated
    for each step during an event for an agent
    """

    def generate_metric(self, params: EpisodeArtifact, **kwargs) -> Metric:

        if "agent_id" not in kwargs:
            raise RuntimeError("Expecting \"agent_id\" to be provided")

        agent_id = kwargs["agent_id"].split('.')[0]

        arr: typing.List[Metric] = []
        for step in params.steps:

            if agent_id not in step.agents or step.agents[agent_id] is None:
                break

            map_obs = step.agents[agent_id].observations

            if map_obs is None:
                continue
            # Create a non terminal metric (Dict) that is comprised of the terminal (Real) observations
            real_dict: typing.Dict[str, Metric] = {key: Real(map_obs[key]) for key in map_obs.keys()}
            arr.append(Dict(real_dict))

        return Vector(arr)


class ControlVector(MetricGeneratorTerminalEventScope):
    """
    Generates vector of Dicts for the control for each
    step during an event for an agent
    """

    def generate_metric(self, params: EpisodeArtifact, **kwargs) -> Metric:

        if "agent_id" not in kwargs:
            raise RuntimeError("Expecting \"agent_id\" to be provided")

        agent_id = kwargs["agent_id"].split('.')[0]

        arr: typing.List[Metric] = []
        for step in params.steps:

            if agent_id not in step.agents or step.agents[agent_id] is None:
                break

            map_act = step.agents[agent_id].actions

            if map_act is None:
                continue
            # Create a non terminal metric (Dict) that is comprised of the terminal (Real) actions
            real_dict: typing.Dict[str, Metric] = {key: Real(map_act[key]) for key in map_act.keys()}
            arr.append(Dict(real_dict))

        return Vector(arr)


class SafetyViolationRatioMetric(MetricGeneratorTerminalEventScope):
    """
    Generates single Real indicating percentage of steps where safety
    is violated. Must use constraint based RTA.
    """

    def generate_metric(self, params: EpisodeArtifact, **kwargs) -> Metric:

        if "agent_id" not in kwargs:
            raise RuntimeError("Expecting \"agent_id\" to be provided")

        agent_id = kwargs["agent_id"].split('.')[0]

        arr = []
        for step in params.steps:

            if agent_id not in step.agents or step.agents[agent_id] is None:
                break

            constraints = step.agents[agent_id].observations['RTAModule']['constraints']

            if constraints is None:
                continue

            arr.append(int(np.any([v < 0 for v in constraints.values()])))

        return Real(np.mean(arr) * 100)


class PositionVelocityVector(MetricGeneratorTerminalEventScope):
    """
    Generates vector of Dicts for the Positions calculated for each
    step during an event for an agent
    """

    def generate_metric(self, params: EpisodeArtifact, **kwargs) -> Metric:

        if "agent_id" not in kwargs:
            raise RuntimeError("Expecting \"agent_id\" to be provided")

        agent_id = kwargs["agent_id"].split('.')[0]
        platform_id = params.agent_to_platforms[agent_id][0]

        arr: typing.List[Metric] = []
        for step in params.steps:

            this_step_platforms = [p['name'] for p in step.platforms]

            if platform_id not in this_step_platforms:
                break

            platform = [p for p in step.platforms if p['name'] == platform_id][0]
            pos = platform['position']
            vel = platform['velocity']

            if pos is None or vel is None:
                continue
            # Create a non terminal metric (Dict) that is comprised of the terminal (Real) observations
            real_dict: typing.Dict[str, Metric] = {'position': pos, 'velocity': vel}
            arr.append(Dict(real_dict))

        return Vector(arr)



"""Reward functions for the docking environment"""

import math
import numpy as np
import safe_autonomy_simulation.entities as e
import safe_autonomy_sims.pettingzoo.docking.utils as utils


def distance_pivot_reward(
    rel_dist: float,
    rel_dist_prev: float,
    c: float = 2.0,
    pivot_ratio: float = 2.0,
    pivot: float = 100,
):
    """A dense exponential reward based on the distance
    between the chief and the deputy.

    The reward is calculated from the current
    and previous state of the system and adjusted
    based on a given pivot value.

    $r_t = c(e^(-ad_t)-e^(ad_(t-1)))$

    where
    * $c$ is a scale factor
    * $a$ is the exponential coefficient given by
    $a = ln(pivot_ratio) / pivot $
    * $d_t$ is the distance between the chief and deputy at time $t$

    The range of the reward is adjusted based on whether the distance
    between the chief and the deputy is greater than or
    less than the pivot value.

    Parameters
    ----------
    rel_dist : float
        current relative distance
    rel_dist_prev : float
        previous relative distance
    c : float, optional
        scale factor, by default 2.0
    pivot_ratio : float, optional
        reward adjustment ratio when d>pivot, by default 2.0
    pivot : int, optional
        distance where the reward gets adjusted, by default 100

    Returns
    -------
    float
        reward value
    """
    a = math.log(pivot_ratio) / pivot
    r = c * (math.exp(-a * rel_dist) - math.exp(-a * rel_dist_prev))
    return r


def delta_v_reward(v: np.ndarray, prev_v: np.ndarray, m: float = 12.0, b: float = 0.0):
    """A dense reward based on the deputy's fuel
    use (change in velocity).

    $r_t = -((\deltav / m) + b)$

    where
    * $\deltav$ is the change in velocity
    * $m$ is the mass of the deputy
    * $b$ is a tunable bias term

    Parameters
    ----------
    v : np.ndarray
        current velocity
    prev_v : np.ndarray
        previous velocity
    m : float, optional
        deputy mass, by default 12.0
    b : float, optional
        bias term, by default 0.0

    Returns
    -------
    float
        reward value
    """
    r = -((utils.delta_v(v=v, prev_v=prev_v) / m) + b)
    return r


def velocity_constraint_reward(v1: np.ndarray, v2: np.ndarray, v_limit: float):
    """A dense reward that punishes the deputy
    for violating a distance-based velocity constraint.

    $r_t = min(-(|v1 - v2| - v_{limit}), 0)$

    where:
    * $v1$ is the deputy's velocity
    * $v2$ is the chief's velocity
    * $v_{limit}$ is a velocity constraint

    Parameters
    ----------
    v1 : np.ndarray
        deputy velocity
    v2 : np.ndarray
        chief velocity
    v_limit : float
        velocity limit

    Returns
    -------
    float
        reward value
    """
    r = min(-(utils.rel_vel(vel1=v1, vel2=v2) - v_limit), 0)
    return r


def docking_success_reward(
    chief: e.PhysicalEntity,
    deputy: e.PhysicalEntity,
    t: float,
    vel_limit: float,
    docking_radius: float = 0.2,
    max_time: float = 2000,
):
    """A sparse reward based on the amount of time
    it took the deputy to successfully dock with
    the chief.

    If docked at a safe velocity:
    $r_t = 1 + (1-t/t_{max})$
    else:
    $r_t = 0$

    Parameters
    ----------
    chief : e.PhysicalEntity
        chief entity
    deputy : e.PhysicalEntity
        deputy entity
    t : float
        current timestep
    vel_limit : float
        maximum safe velocity
    docking_radius : float, optional
        radius of the docking region, by default 0.2
    max_time : int, optional
        maximum allowed time to dock, by default 2000

    Returns
    -------
    float
        reward value
    """
    in_docking = (
        utils.rel_dist(pos1=chief.position, pos2=deputy.position) < docking_radius
    )
    safe_velocity = utils.rel_vel(vel1=chief.velocity, vel2=deputy.velocity) < vel_limit
    if in_docking and safe_velocity:
        r = 1.0 + (1 - t / max_time)
    else:
        r = 0
    return r


def timeout_reward(t: float, max_time: float = 2000):
    """A sparse reward that punishes the agent
    for not completing the task within the time
    limit.

    Parameters
    ----------
    t : float
        current simulation time
    max_time : float, optional
        maximum time allowed to complete the task, by default 2000

    Returns
    -------
    float
        reward value
    """
    r = 0
    if t > max_time:
        r = -1.0
    return r


def crash_reward(
    chief: e.PhysicalEntity,
    deputy: e.PhysicalEntity,
    vel_limit: float,
    docking_radius: float = 0.2,
):
    """A sparse reward that punishes the agent
    for entering the docking region at an unsafe velocity
    (crashing).

    Parameters
    ----------
    chief : e.PhysicalEntity
        chief entity
    deputy : e.PhysicalEntity
        deputy entity
    vel_limit : float
        maximum safe velocity
    docking_radius : float, optional
        radius of the docking region, by default 0.2

    Returns
    -------
    float
        reward value
    """
    in_docking = (
        utils.rel_dist(pos1=chief.position, pos2=deputy.position) < docking_radius
    )
    safe_velocity = utils.rel_vel(vel1=chief.velocity, vel2=deputy.velocity) < vel_limit
    if in_docking and not safe_velocity:
        r = -1.0
    else:
        r = 0
    return r


def out_of_bounds_reward(
    chief_pos: np.ndarray, deputy_pos: np.ndarray, max_distance: float = 10000
):
    """A sparse reward that punishes the agent
    for going out of bounds.

    Parameters
    ----------
    chief_pos : np.ndarray
        chief position
    deputy_pos : np.ndarray
        deputy position
    max_distance : int, optional
        maximum allowed distance from the chief, by default 10000

    Returns
    -------
    float
        reward value
    """
    d = utils.rel_dist(pos1=chief_pos, pos2=deputy_pos)
    if d > max_distance:
        r = -1.0
    else:
        r = 0
    return r



"""
--------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning Core (CoRL) Safe Autonomy Extension.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------

This module contains custom illumination functions utilized in the CWH inspection task.
"""

import math
from csv import writer

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.patches import Circle

########################################################## ILLUMINATION ##########################################################


def get_sun_position(current_time, angular_velocity, initial_theta, r_avg):
    """
    Returns current sun position based on initial sun position and elapsed time

    Parameters
    ----------
    current_time: int
        current simulation time in seconds
    angular_velocity: float
        mean motion of sun in meters per second
    initial_theta: float
        initial angle of sun with respect to chief in radians
    r_avg: float
        average distance from earth to sun in meters

    Returns
    -------
    sun_position: array
        array of sun position in meters (CWH coordinates)
    """
    d_theta = angular_velocity
    current_theta = d_theta * current_time + initial_theta
    sun_position = [r_avg * (math.cos(current_theta)), -r_avg * (math.sin(current_theta)), 0]

    return sun_position


def check_illum(point, sun_angle, r_avg, radius):
    """
    Receive a candidate point as an input
    Receive sun position as an input
    Output should be boolean (sun obstructed or not)

    Parameters
    ----------
    point: array
        point on chief
    sun_angle: float
        current sun angle
    r_avg: float
        average distance from earth to sun in meters
    radius: float
        radius of chief in meters

    Returns
    -------
    bool_val: bool
        boolean assigned for illuminated or not
    """

    # Chief position is origin [cwh dynamics]
    center = [0, 0, 0]
    normal_to_surface = normalize(point)
    # Get a point slightly off the surface of the sphere so don't detect surface as an intersection
    shifted_point = point + 1e-5 * normal_to_surface
    sun_position = [r_avg * (math.cos(sun_angle)), -r_avg * (math.sin(sun_angle)), 0]
    intersection_to_light = normalize(sun_position - shifted_point)

    intersect_var = sphere_intersect(center, radius, shifted_point, intersection_to_light)

    bool_val = False
    # No intersection means that the point in question is illuminated in some capacity
    # (i.e. the point on the chief is not blocked by the chief itself)
    if intersect_var is None:
        bool_val = True

    return bool_val


def num_inspected_points(points):
    """
    Returns number of inspected points

    Parameters
    ----------
    points : dict
        set of inspection points
    """
    pts = 0
    for point in points:
        if points[point]:
            pts += 1

    return pts


def save_data(points, current_time, position, sun_position, action, velocity, path):
    """Save environment data

    Parameters
    ----------
    points : dict
        inspection points
    current_time : float
        current sim time
    position : list
        position vector
    sun_position : list
        sun position vector
    action : list
        action vector
    velocity : list
        velocity vector
    path : str
        path to save data
    """
    points_bool = []
    for point in points:
        if not points[point]:
            points_bool.append(0)
        else:
            points_bool.append(1)

    # Write to csv
    temp = [current_time, position, points_bool, sun_position, action, velocity]
    with open(path, 'a') as f_object:
        writer_object = writer(f_object)
        writer_object.writerow(temp)
        f_object.close()


def evaluate_RGB(RGB):
    """
    Receive RGB array
    Return boolean based on thresholding logic
    For now, only would work for red

    Parameters
    ----------
    RGB: array
        3x1 array containing RGB value

    Returns
    -------
    RGB_bool: bool
        boolean assigned for illuminated or not
    """
    RGB_bool = True

    # Too dark
    if RGB[0] < .12:
        RGB_bool = False
    # Too bright/white
    if RGB[0] > .8 and RGB[1] > .8 and RGB[2] > .8:
        RGB_bool = False

    return RGB_bool


def compute_illum_pt(point, sun_angle, deputy_position, r_avg, radius, chief_properties, light_properties):
    """
    Receive a candidate point as an input
    Receive sun position as an input
    Returns a color (all zeros (black) for not illuminated)
    """
    # Chief position is origin [cwh dynamics]
    center = [0, 0, 0]
    normal_to_surface = normalize(point)
    # Get a point slightly off the surface of the sphere so don't detect surface as an intersection
    shifted_point = point + 1e-5 * normal_to_surface
    sun_position = [r_avg * (math.cos(sun_angle)), -r_avg * (math.sin(sun_angle)), 0]
    intersection_to_light = normalize(sun_position - shifted_point)
    intersect_var = sphere_intersect(center, radius, shifted_point, intersection_to_light)

    illumination = np.zeros((3))
    # No intersection means that the point in question is illuminated in some capacity
    # (i.e. the point on the chief is not blocked by the chief itself)
    if intersect_var is None:
        # Blinn-Phong Illumination Model
        # https://en.wikipedia.org/wiki/Blinn%E2%80%93Phong_reflection_model
        illumination += np.array(chief_properties['ambient']) * np.array(light_properties['ambient'])
        illumination += np.array(chief_properties['diffuse']) * np.array(light_properties['diffuse']) * \
            np.dot(intersection_to_light, normal_to_surface)
        intersection_to_camera = normalize(deputy_position - point)
        H = normalize(intersection_to_light + intersection_to_camera)
        illumination += np.array(chief_properties['specular']) * np.array(light_properties['specular']) * \
            np.dot(normal_to_surface, H)**(chief_properties['shininess'] / 4)
        illumination = np.clip(illumination, 0, 1)
    return illumination


def save_image_test_delete(color):
    """
    Recieves RGB array and saves an image
    """
    image = np.zeros((200, 200, 3))
    for i in range(200):
        for j in range(200):
            image[i, j] = color
    string = str(round(color[0], 2)) + '_' + str(round(color[1], 2)) + '_' + str(round(color[2], 2))
    string = string.replace('.', ',')
    string = string + '.png'
    plt.imsave('figs_results/' + string, image)


def compute_illum(deputy_position, sun_position, resolution, radius, focal_length, chief_properties, light_properties, pixel_pitch):
    # pylint: disable-msg=too-many-locals
    """
    Renders the full scene using backwards ray tracing and returns a full RGB image
    """
    ratio = float(resolution[0]) / resolution[1]
    # For now, assuming deputy sensor always pointed at chief (which is origin)
    chief_position = [0, 0, 0]
    sensor_dir = normalize(chief_position - deputy_position)
    image_plane_position = deputy_position + sensor_dir * focal_length
    # There are an infinite number of vectors normal to sensor_dir -- choose one
    x = -1
    y = 1
    z = -(image_plane_position[0] * x + image_plane_position[1] * y) / image_plane_position[2]
    norm1 = normalize([x, y, z])

    # np.cross bug work-around https://github.com/microsoft/pylance-release/issues/3277
    def cross2(a: np.ndarray, b: np.ndarray) -> np.ndarray:
        return np.cross(a, b)

    norm2 = cross2(sensor_dir, norm1)
    # Used for x,y,z pixel locations - there will be resolution[0] * resolution[1] pixels
    x_width = np.tan((pixel_pitch / focal_length) / 2) * 2 * focal_length
    norm1_range = x_width
    norm2_range = x_width / ratio
    step_norm1 = norm1_range / (resolution[0])
    step_norm2 = norm2_range / (resolution[1])
    # 3D matrix (ie. height-by-width matrix with each entry being an array of size 3) which creates an image
    image = np.zeros((resolution[1], resolution[0], 3))
    for i in range(int(resolution[1])):  # y coords
        for j in range(int(resolution[0])):  # x coords
            # Initialize pixel
            illumination = np.zeros((3))
            # Convert to CWH coordinates
            pixel_location = image_plane_position + ((norm2_range / 2) -
                                                     (i * step_norm2)) * (norm2) + (-(norm1_range / 2) + (j * step_norm1)) * (norm1)
            ray_direction = normalize(pixel_location - deputy_position)
            dist_2_intersect = sphere_intersect(chief_position, radius, deputy_position, ray_direction)
            # Light ray hits sphere, so we continue - else get next pixel
            if dist_2_intersect is not None:
                intersection_point = deputy_position + dist_2_intersect * ray_direction
                normal_to_surface = normalize(intersection_point - chief_position)
                shifted_point = intersection_point + 1e-5 * normal_to_surface
                intersection_to_light = normalize(sun_position - shifted_point)
                intersect_var = sphere_intersect(chief_position, radius, shifted_point, intersection_to_light)
                # If the shifted point doesn't intersect with the chief on the way to the light, it is unobstructed
                if intersect_var is None:
                    # Blinn-Phong Illumination Model
                    # https://en.wikipedia.org/wiki/Blinn%E2%80%93Phong_reflection_model

                    illumination += np.array(chief_properties['ambient']) * np.array(light_properties['ambient'])
                    illumination += np.array(chief_properties['diffuse']) * np.array(light_properties['diffuse']) * \
                        np.dot(intersection_to_light, normal_to_surface)
                    intersection_to_camera = normalize(deputy_position - intersection_point)
                    H = normalize(intersection_to_light + intersection_to_camera)
                    illumination += np.array(chief_properties['specular']) * np.array(light_properties['specular']) * \
                        np.dot(normal_to_surface, H)**(chief_properties['shininess'] / 4)
                # Shadowed
                else:
                    continue
            image[i, j] = np.clip(illumination, 0, 1)
    return image


def normalize(vector):
    """
    Normalize
    """
    return vector / np.linalg.norm(vector)


def sphere_intersect(center, radius, ray_origin, ray_direction):
    """
    Sphere intersection, returns closest distance if intersection found,
    Returns None upon no intersection
    """
    b = 2 * np.dot(ray_direction, ray_origin - center)
    c = np.linalg.norm(ray_origin - center)**2 - radius**2
    delta = b**2 - 4 * c
    if delta > 0:
        t1 = (-b + np.sqrt(delta)) / 2
        t2 = (-b - np.sqrt(delta)) / 2
        if t1 > 0 and t2 > 0:
            return min(t1, t2)
    # No intersection
    return None


########################################################## ILLUMINATION ##########################################################

################################################# VISUALIZATION AND VERIFICATION #################################################


def visualizePoints(deputy_position, current_time, sun_position, points):
    """
    Plot the points on the chief and light up green when inspected
    Saves a figure locally
    """
    fig = plt.figure(1)
    ax = fig.gca(projection='3d')
    ax.set_title("3D Plot at time: " + str(current_time) + " (sec)")
    plt.grid()
    # Change resolution with this and changing the dpi
    fig.set_size_inches(8, 8)
    chief_position = [0, 0, 0]
    sun_vector = normalize(np.array(sun_position) - chief_position)
    line_scalar = 200
    point_inSunDir = chief_position + line_scalar * sun_vector
    ax.plot(
        [point_inSunDir[0], chief_position[0]], [point_inSunDir[1], chief_position[1]], [point_inSunDir[2], chief_position[2]],
        color='#FFD700',
        linewidth=6,
        label='Sun vector'
    )
    temp1 = 0
    temp2 = 0
    for point in points:
        if points[point]:
            if temp1 == 0:
                ax.scatter3D(point[0], point[1], point[2], marker='.', color='green', label='Inspected Point')
                temp1 += 1
            else:
                ax.scatter3D(point[0], point[1], point[2], marker='.', color='green')
        else:
            if temp2 == 0:
                ax.scatter3D(point[0], point[1], point[2], marker='.', color='red', label='Uninspected Point')
                temp2 += 1
            else:
                ax.scatter3D(point[0], point[1], point[2], marker='.', color='red')

    ax.scatter3D(deputy_position[0], deputy_position[1], deputy_position[2], marker='o', color='blue', label='Deputy Spacecraft')
    ax.set_xlabel('X in CWH C.S. [meters]')
    ax.set_ylabel('Y in CWH C.S. [meters]')
    ax.set_zlabel('Z in CWH C.S. [meters]')
    ax.set_box_aspect((1, 1, 1))
    set_axes_equal(ax)
    ax.legend()
    plt.savefig('figs_results/' + 'timestep_' + str(current_time) + 'sec_PointPlot.png', dpi=100)
    plt.close(fig)


def visualize3D(deputy_position, current_time, sun_position, radius):
    """
    Plot deputy, chief and sun direction vector
    Saves a figure locally
    """
    fig = plt.figure(1)
    ax = fig.gca(projection='3d')
    ax.set_title("3D Plot at time: " + str(current_time) + " (sec)")
    plt.grid()
    # Change resolution with this and changing the dpi
    fig.set_size_inches(8, 8)
    chief_position = [0, 0, 0]
    sun_vector = normalize(np.array(sun_position) - chief_position)
    line_scalar = 200
    point_inSunDir = chief_position + line_scalar * sun_vector
    ax.plot(
        [point_inSunDir[0], chief_position[0]], [point_inSunDir[1], chief_position[1]], [point_inSunDir[2], chief_position[2]],
        color='#FFD700',
        linewidth=6,
        label='Sun vector'
    )
    u, v = np.mgrid[0:2 * np.pi:30j, 0:np.pi:20j]
    x = radius * np.cos(u) * np.sin(v)
    y = radius * np.sin(u) * np.sin(v)
    z = radius * np.cos(v)
    ax.plot_surface(x, y, z, color='red')
    ax.scatter3D(deputy_position[0], deputy_position[1], deputy_position[2], marker='o', color='blue', label='Deputy Spacecraft')
    ax.set_xlabel('X in CWH C.S. [meters]')
    ax.set_ylabel('Y in CWH C.S. [meters]')
    ax.set_zlabel('Z in CWH C.S. [meters]')
    ax.set_box_aspect((1, 1, 1))
    set_axes_equal(ax)
    ax.legend()
    plt.savefig('figs_results/' + 'timestep_' + str(current_time) + 'sec_3DPlot.png', dpi=100)
    plt.close(fig)


def save_image(RGB, current_time):
    """
    Recieves RGB array and saves an image
    """

    string = 'timestep_' + str(current_time) + 'sec_camera.png'
    plt.imsave('figs_results/' + string, RGB)


def concat_images(imga, imgb):
    """
    Combines two color image ndarrays side-by-side.
    """
    ha, wa = imga.shape[:2]
    hb, wb = imgb.shape[:2]
    max_height = np.max([ha, hb])
    total_width = wa + wb
    new_img = np.ones(shape=(max_height, total_width, 3))

    if max_height == hb:
        delta = (int(max_height / 2) - int(ha / 2))
        new_img[delta:delta + ha, :wa] = imga
        new_img[:hb, wa:wa + wb] = imgb
    if max_height == ha:
        delta = (int(max_height / 2) - int(hb / 2))
        new_img[:ha, :wa] = imga
        new_img[delta:delta + hb, wa:wa + wb] = imgb
    return new_img


def concat_n_images(image_path_list):
    """
    Combines N color images from a list of image paths.
    """
    output = None
    for i, img_path in enumerate(image_path_list):
        img = plt.imread(img_path)[:, :, :3]
        if i == 0:
            output = img
        else:
            output = concat_images(output, img)
    return output


def combine_images(image1, image2, current_time):
    """
    Combines images into one image and saves result locally
    """
    im_array = [image1, image2]
    final_image = concat_n_images(im_array)
    string = 'timestep_' + str(current_time) + 'sec_COMBINED.png'
    plt.imsave('figs_results/' + string, final_image)


def render_subplots(fig, axes, deputy_position, sun_position, radius, current_time, step_rate):
    # pylint: disable-msg=too-many-statements
    """
    Real-time rendering of scene with subplots (xy, xz, yz)
    TODO: Fix autoscaling problem (need equal axes so sphere looks normal)
    """
    ax_3d = axes[0]
    ax_xy = axes[1]
    ax_xz = axes[2]
    ax_yz = axes[3]
    line_scalar = 200
    chief_position = [0, 0, 0]
    sun_vector = normalize(np.array(sun_position) - chief_position)
    point_inSunDir = chief_position + line_scalar * sun_vector
    # Only runs once at first step
    if current_time == step_rate:
        fig.suptitle('Real-time 3D Inspection Problem with Illumination', fontsize=16)
        ax_3d.set(xlabel='X in CWH C.S. [meters]', ylabel='Y in CWH C.S. [meters]', zlabel='Z in CWH C.S. [meters]')
        # Subplots
        ax_xy.set(xlabel='X in CWH C.S. [meters]', ylabel='Y in CWH C.S. [meters]')
        ax_xz.set(xlabel='X in CWH C.S. [meters]', ylabel='Z in CWH C.S. [meters]')
        ax_yz.set(xlabel='Y in CWH C.S. [meters]', ylabel='Z in CWH C.S. [meters]')
        fig.set_size_inches(10, 10)
        ax_3d.view_init(elev=20, azim=56)
        u, v = np.mgrid[0:2 * np.pi:30j, 0:np.pi:20j]
        x = radius * np.cos(u) * np.sin(v)
        y = radius * np.sin(u) * np.sin(v)
        z = radius * np.cos(v)
        ax_3d.plot_surface(x, y, z, color='red')
        ax_3d.scatter3D(deputy_position[0], deputy_position[1], deputy_position[2], marker='o', color='blue', label='Deputy Spacecraft')
        ax_3d.scatter3D(chief_position[0], chief_position[1], chief_position[2], marker='o', color='red', label='Chief Spacecraft')
        ax_3d.plot(
            [point_inSunDir[0], chief_position[0]], [point_inSunDir[1], chief_position[1]], [point_inSunDir[2], chief_position[2]],
            color='#FFD700',
            linewidth=6,
            label='Sun vector'
        )
        ax_3d.set_box_aspect((1, 1, 1))
        set_axes_equal(ax_3d)
        ax_3d.legend()
        circle1 = Circle((0, 0), radius, color='red')
        circle2 = Circle((0, 0), radius, color='red')
        circle3 = Circle((0, 0), radius, color='red')
        ax_xy.scatter(deputy_position[0], deputy_position[1], marker='o', color='blue')
        ax_xy.add_patch(circle1)
        ax_xz.scatter(deputy_position[0], deputy_position[2], marker='o', color='blue')
        ax_xz.add_patch(circle2)
        ax_yz.scatter(deputy_position[1], deputy_position[2], marker='o', color='blue')
        ax_yz.add_patch(circle3)
        ax_xz.axis('square')
        ax_xy.axis('square')
        ax_yz.axis('square')
        plt.ion()
        plt.show()
    else:
        ax_3d.scatter3D(deputy_position[0], deputy_position[1], deputy_position[2], marker='o', color='blue', label='Deputy Spacecraft')
        ax_3d.plot(
            [point_inSunDir[0], chief_position[0]], [point_inSunDir[1], chief_position[1]], [point_inSunDir[2], chief_position[2]],
            color='#FFD700',
            linewidth=6,
            label='Sun vector'
        )
        ax_3d.autoscale()
        ax_xy.scatter(deputy_position[0], deputy_position[1], marker='o', color='blue')
        ax_xz.scatter(deputy_position[0], deputy_position[2], marker='o', color='blue')
        ax_yz.scatter(deputy_position[1], deputy_position[2], marker='o', color='blue')
        ax_xz.axis('square')
        ax_xy.axis('square')
        ax_yz.axis('square')
    plt.draw()
    plt.pause(.0001)


def render_3d(fig, deputy_position, sun_position, radius, current_time, step_rate):
    """
    Real-time rendering of scene
    TODO: Fix autoscaling problem (need equal axes so sphere looks normal)
    """
    # ax = fig.gca(projection='3d')
    ax = fig.add_subplot(projection='3d')
    line_scalar = 200
    # Initialization
    if current_time == step_rate:
        fig.set_size_inches(10, 10)
        ax.set_title("3D Plot, Inspection Problem")
        ax.set_xlabel('X in CWH C.S. [meters]')
        ax.set_ylabel('Y in CWH C.S. [meters]')
        ax.set_zlabel('Z in CWH C.S. [meters]')
        ax.view_init(elev=20, azim=56)
        u, v = np.mgrid[0:2 * np.pi:30j, 0:np.pi:20j]
        x = radius * np.cos(u) * np.sin(v)
        y = radius * np.sin(u) * np.sin(v)
        z = radius * np.cos(v)
        ax.plot_surface(x, y, z, color='red')
        ax.scatter3D(deputy_position[0], deputy_position[1], deputy_position[2], marker='o', color='blue', label='Deputy Spacecraft')
        chief_position = [0, 0, 0]
        ax.scatter3D(chief_position[0], chief_position[1], chief_position[2], marker='o', color='red', label='Chief Spacecraft')
        sun_vector = normalize(np.array(sun_position) - chief_position)
        point_inSunDir = chief_position + line_scalar * sun_vector
        ax.plot(
            [point_inSunDir[0], chief_position[0]], [point_inSunDir[1], chief_position[1]], [point_inSunDir[2], chief_position[2]],
            color='#FFD700',
            linewidth=6,
            label='Sun vector'
        )
        ax.set_box_aspect((1, 1, 1))
        set_axes_equal(ax)
        ax.legend()
        plt.ion()
        plt.show()
    else:
        ax.scatter3D(deputy_position[0], deputy_position[1], deputy_position[2], marker='o', color='blue', label='Deputy Spacecraft')
        chief_position = [0, 0, 0]
        sun_vector = normalize(np.array(sun_position) - chief_position)
        point_inSunDir = chief_position + line_scalar * sun_vector
        ax.plot(
            [point_inSunDir[0], chief_position[0]], [point_inSunDir[1], chief_position[1]], [point_inSunDir[2], chief_position[2]],
            color='#FFD700',
            linewidth=6,
            label='Sun vector'
        )
        ax.autoscale()

    plt.draw()
    plt.pause(.0001)


def set_axes_equal(ax: plt.Axes):
    """Set 3D plot axes to equal scale.

    Make axes of 3D plot have equal scale so that spheres appear as
    spheres and cubes as cubes.  Required since `ax.axis('equal')`
    and `ax.set_aspect('equal')` don't work on 3D.
    """
    limits = np.array([
        ax.get_xlim3d(),
        ax.get_ylim3d(),
        ax.get_zlim3d(),
    ])

    origin = np.mean(limits, axis=1)
    radius = 0.5 * np.max(np.abs(limits[:, 1] - limits[:, 0]))
    _set_axes_radius(ax, origin, radius)


def _set_axes_radius(ax, origin, radius):
    x, y, z = origin
    ax.set_xlim3d([x - radius, x + radius])
    ax.set_ylim3d([y - radius, y + radius])
    ax.set_zlim3d([z - radius, z + radius])


################################################# VISUALIZATION AND VERIFICATION #################################################



"""
--------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core  Extension.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------

This module defines an experiment class which runs training episodes via ray rllib's API.
This is useful for debugging and validating new environments and can easily be modified to log state data.
"""
# pylint: disable=C0301,R0914,R1732,W0612,R0915

import argparse
import pickle
import socket
import typing
from datetime import datetime

import ray
import tqdm
from corl.environment.multi_agent_env import ACT3MultiAgentEnv, ACT3MultiAgentEnvValidator
from corl.episode_parameter_providers.remote import RemoteEpisodeParameterProvider
from corl.experiments.rllib_experiment import RllibExperiment, RllibExperimentValidator
from corl.libraries.factory import Factory

# import ray.rllib.agents.ppo as ppo
from ray.rllib.algorithms.ppo import ppo


class RllibAPIExperimentValidator(RllibExperimentValidator):
    """
    The validator for the RllibAPIExperiment class.

    Attributes
    ----------
    serialized_ray_config_path: str
        path to params.pkl file
    trained_models_checkpoint_path: str
        path to checkpoint directory
    """
    serialized_ray_config_path: str
    trained_models_checkpoint_path: str


class RllibAPIExperiment(RllibExperiment):
    """
    This class loads an environment from given checkpoint folders, then runs 10 rollout
    training episdoes using rllib's API directly. This allows for debugging with access to state, done,
    and reward data and logic for environment development and validation.
    """

    @staticmethod
    def get_validator() -> typing.Type[RllibAPIExperimentValidator]:
        return RllibAPIExperimentValidator

    def run_experiment(self, args: argparse.Namespace):

        # setup
        rllib_config = self._select_rllib_config(args.compute_platform)

        if args.compute_platform in ['ray']:
            self._update_ray_config_for_ray_platform()

        self._add_trial_creator()

        # This needs to be before the ray cluster is initialized
        if args.debug:
            self.config.ray_config['local_mode'] = True

        ray.init(**self.config.ray_config)

        ray_resources = ray.available_resources()

        self.config.env_config["agents"], self.config.env_config["agent_platforms"] = self.create_agents(
            args.platform_config, args.agent_config
        )

        self.config.env_config["horizon"] = rllib_config["horizon"]

        if args.other_platform:
            self.config.env_config["other_platforms"] = self.create_other_platforms(args.other_platform)

        if not self.config.ray_config['local_mode']:
            self.config.env_config['episode_parameter_provider'] = RemoteEpisodeParameterProvider.wrap_epp_factory(
                Factory(**self.config.env_config['episode_parameter_provider']),
                actor_name=ACT3MultiAgentEnv.episode_parameter_provider_name
            )

            for agent_name, agent_configs in self.config.env_config['agents'].items():
                agent_configs.class_config.config['episode_parameter_provider'] = RemoteEpisodeParameterProvider.wrap_epp_factory(
                    Factory(**agent_configs.class_config.config['episode_parameter_provider']), agent_name
                )

        self.config.env_config['epp_registry'] = ACT3MultiAgentEnvValidator(**self.config.env_config).epp_registry

        tmp = ACT3MultiAgentEnv(self.config.env_config)
        tmp_as = tmp.action_space
        tmp_os = tmp.observation_space
        tmp_ac = self.config.env_config['agents']

        policies = {
            policy_name: (
                tmp_ac[policy_name].policy_config["policy_class"],
                policy_obs,
                tmp_as[policy_name],
                tmp_ac[policy_name].policy_config["config"]
            )
            for policy_name,
            policy_obs in tmp_os.spaces.items()
            if tmp_ac[policy_name]
        }

        train_policies = [policy_name for policy_name in policies.keys() if tmp_ac[policy_name].policy_config["train"]]

        rllib_config["multiagent"] = {
            "policies": policies, "policy_mapping_fn": lambda agent_id: agent_id, "policies_to_train": train_policies
        }

        rllib_config["env"] = ACT3MultiAgentEnv
        rllib_config["callbacks"] = self.get_callbacks()
        rllib_config["env_config"] = self.config.env_config
        now = datetime.now()
        rllib_config["env_config"]["output_date_string"] = f"{now.strftime('%Y%m%d_%H%M%S')}_{socket.gethostname()}"
        rllib_config["create_env_on_driver"] = True
        rllib_config["batch_mode"] = "complete_episodes"

        if args.debug:
            rllib_config['num_workers'] = 0

        self._enable_episode_parameter_provider_checkpointing()

        if args.profile:
            if "stop" not in self.config.tune_config:
                self.config.tune_config["stop"] = {}
            self.config.tune_config["stop"]["training_iteration"] = args.profile_iterations

        search_class = None
        if self.config.hparam_search_class is not None:
            if self.config.hparam_search_config is not None:
                search_class = self.config.hparam_search_class(**self.config.hparam_search_config)
            else:
                search_class = self.config.hparam_search_class()
            search_class.add_algorithm_hparams(rllib_config, self.config.tune_config)

        ################################# Manual rllib API Episodes ####################################

        # instantiate env
        env = ACT3MultiAgentEnv(self.config.env_config)

        # load config from checkpoint
        ray_config_file = open(self.config.serialized_ray_config_path, "rb")
        ray_config = pickle.load(ray_config_file)

        # override serialized environment config
        ray_config["env_config"] = self.config.env_config

        # instantiate agents and restore models from checkpoint
        agents = ppo.PPOTrainer(config=ray_config, env=rllib_config['env'])
        agents.restore(self.config.trained_models_checkpoint_path)

        # step for loop
        num_rollouts = 10
        for _ in tqdm.tqdm(range(num_rollouts)):
            # run until episode ends
            dones = {"placeholder": False}
            obs = env.reset()
            step_num = 0

            while not all(dones.values()):
                # progress environment state
                action_dict = {}
                for name in obs.keys():
                    action_dict[name] = agents.compute_single_action(obs[name], policy_id=name)

                obs, rewards, dones, info = env.step(action_dict)
                step_num += 1
